import os
import copy
import numpy as np
import pandas as pd

from mpunet.utils import create_folders


def init_result_dict_3D(all_images, n_classes):
    if n_classes == 1:
        n_classes = 2

    # Prepare dictionary of per class results
    detailed_results = {"class": range(1, n_classes)}
    for image_id in all_images:
        detailed_results.update({image_id: [np.nan for _ in range(1, n_classes)]})

    return {image_id: [np.nan] for image_id in all_images}, detailed_results


def save_all_3D(results, detailed_res, out_dir):
    txt_res_dir = os.path.join(out_dir, "txt")
    csv_res_dir = os.path.join(out_dir, "csv")

    # Create folders
    create_folders([txt_res_dir, csv_res_dir])

    # Save main text/csv results files
    results_to_txt(results, txt_res_dir, transpose=True)
    results_to_csv(results, csv_res_dir, transpose=True)
    results_to_txt(detailed_res, txt_res_dir, fname="detailed")
    results_to_csv(detailed_res, csv_res_dir, fname="detailed")


def init_result_dicts(views, all_images, n_classes):
    if n_classes == 1:
        n_classes = 2

    results = {"identifier": [image_id for image_id in sorted(all_images)]}
    results.update({str(v): [np.nan for _ in all_images] for v in views})
    results.update({"MJ": [np.nan for _ in all_images]})
    results = pd.DataFrame(results)
    results = results.set_index("identifier")

    # Prepare dictionary of per class results
    inner = {"class": range(1, n_classes)}
    for image_id in all_images:
        inner.update({image_id: [np.nan for _ in range(1, n_classes)]})
    views = list(views) + ["MJ"]
    pc_results = {str(v): pd.DataFrame(copy.deepcopy(inner)) for v in views}
    pc_results = {key: df.set_index("class") for key, df in pc_results.items()}
    # pc_results.update({"MJ": pd.DataFrame(copy.deepcopy(inner))})

    return results, pc_results


def load_result_dicts(csv_dir, views):
    from glob import glob
    import re
    regex = re.compile(r"[-]?\d{1}[.]\d+")

    def match_view_and_file(view, path):
        fname = os.path.splitext(os.path.split(path)[-1])[0]
        path_view_components = np.array(re.findall(regex, fname), np.float)
        if len(path_view_components) == 0:
            # MJ.csv or results.csv
            return False
        assert len(path_view_components) == 3
        assert len(view) == 3
        return np.all(path_view_components.round(4) == view.round(4))

    csv_dir = os.path.abspath(csv_dir)
    pc_results = {"MJ": pd.read_csv(csv_dir + "/MJ.csv", index_col=0)}
    results = pd.read_csv(csv_dir + "/results.csv", index_col=0)

    paths = [p for p in glob(csv_dir + "/*csv")]
    for v in views:
        found_match = False
        for path in paths:
            if match_view_and_file(v, path):
                pc_results[str(v)] = pd.read_csv(path, index_col=0)
                found_match = True
        if not found_match:
            raise RuntimeError(
                "Could not infer relationship between view {} and view "
                "csv files")
    return results, pc_results


def to_df(results, transpose=False):
    df = pd.DataFrame(results)
    try:
        df = df.set_index("class")
    except KeyError:
        pass
    if transpose:
        df = df.transpose()
    return df


def results_to_csv(results, res_path, fname=None, transpose=False):
    if fname is None:
        fname = "results"

    # Save results
    df = to_df(results, transpose)
    with open(os.path.join(res_path, "%s.csv" % fname), "w") as out_file:
        out_file.write(df.to_csv(index=True) + "\n")


def results_to_txt(results, res_path, fname=None, transpose=False):
    if fname is None:
        fname = "results"

    # Save results
    df = to_df(results, transpose)
    with open(os.path.join(res_path, "%s.txt" % fname), "w") as out_file:
        out_file.write(df.to_string() + "\n")


def save_all(results, pc_results, out_dir):

    # Get output paths
    txt_res_dir = os.path.join(out_dir, "txt")
    csv_res_dir = os.path.join(out_dir, "csv")

    # Create folders
    create_folders([txt_res_dir, csv_res_dir])

    # Save main text/csv results files
    results_to_txt(results, txt_res_dir)
    results_to_csv(results, csv_res_dir)

    # Write detailed results
    for view in pc_results:
        r = pc_results[view]
        view_str = str(view).replace("[", "").strip().replace("]", "").replace(" ", "_")
        results_to_txt(r, txt_res_dir, fname=view_str)
        results_to_csv(r, csv_res_dir, fname=view_str)



from mpunet.image.queue.base_queue import BaseQueue
from contextlib import contextmanager


class EagerQueue(BaseQueue):
    """
    Implements a queue-like object (same API interface as LoadQueue), but one
    that loads all data immediately when initialized.
    This is useful for wrapping a smaller collection of data in an object that
    behaves similar to the training queue object, but where all data is loaded
    up-front.
    """
    def __init__(self, dataset, logger=None, **kwargs):
        """
        TODO
        Args:
            dataset:
            logger:
        """
        super(EagerQueue, self).__init__(
            dataset=dataset,
            logger=logger
        )
        self.logger("'Eager' queue created:\n"
                    "  Dataset:      {}".format(self.dataset))
        self.logger("Preloading all {} images now... (eager)".format(len(dataset)))
        self.dataset.load()

    @staticmethod
    def check_loaded(image):
        if not image.is_loaded:
            raise RuntimeError("Some process unloaded image '{}'; this "
                               "is unexpected behaviour when using the "
                               "EagerQueue object, which expects all data to "
                               "be loaded at all times")
        return image

    def __iter__(self):
        for i in range(len(self.dataset.images)):
            with self.get_image_by_idx(i) as ss:
                yield ss

    @contextmanager
    def get_random_image(self):
        yield self.check_loaded(super().get_random_image())

    @contextmanager
    def get_image_by_idx(self, image_idx):
        yield self.check_loaded(super().get_image_by_idx(image_idx))

    @contextmanager
    def get_image_by_id(self, image_id):
        yield self.check_loaded(super().get_image_by_id(image_id))



import numpy as np
from mpunet.interpolation import RegularGridInterpolator
from scipy.ndimage.filters import gaussian_filter


def elastic_transform_2d(image, labels, alpha, sigma, bg_val=0.0):
    """
    Elastic deformation of images as described in [Simard2003]_.
    [Simard2003] Simard, Steinkraus and Platt, "Best Practices for
    Convolutional Neural Networks applied to Visual Document Analysis", in
    Proc. of the International Conference on Document Analysis and
    Recognition, 2003.

    Modified from:
    https://gist.github.com/chsasank/4d8f68caf01f041a6453e67fb30f8f5a

    Modified to take 3 and 4 dimensional inputs
    Deforms both the image and corresponding label file
    image tri-linear interpolated
    Label volumes nearest neighbour interpolated
    """
    if image.ndim == 2:
        image = np.expand_dims(image, axis=-1)
    shape = image.shape[:2]
    channels = image.shape[-1]
    dtype = image.dtype
    bg_val = bg_val if isinstance(bg_val, (list, tuple, np.ndarray)) \
        else [bg_val] * channels

    # Define coordinate system
    coords = np.arange(shape[0]), np.arange(shape[1])

    # Initialize interpolators
    im_intrps = []
    for i in range(channels):
        im_intrps.append(RegularGridInterpolator(coords, image[..., i],
                                                 method="linear",
                                                 bounds_error=False,
                                                 fill_value=bg_val[i],
                                                 dtype=np.float32))

    # Get random elastic deformations
    dx = gaussian_filter((np.random.rand(*shape) * 2 - 1), sigma,
                         mode="constant", cval=0.) * alpha
    dy = gaussian_filter((np.random.rand(*shape) * 2 - 1), sigma,
                         mode="constant", cval=0.) * alpha

    # Define sample points
    x, y = np.mgrid[0:shape[0], 0:shape[1]]
    indices = np.reshape(x + dx, (-1, 1)), \
              np.reshape(y + dy, (-1, 1))

    # Interpolate all image channels
    image = np.empty(shape=image.shape, dtype=dtype)
    for i, intrp in enumerate(im_intrps):
        image[..., i] = intrp(indices).reshape(shape)

    # Interpolate labels
    if labels is not None:
        lab_intrp = RegularGridInterpolator(coords, labels,
                                            method="nearest",
                                            bounds_error=False,
                                            fill_value=0,
                                            dtype=np.uint8)

        labels = lab_intrp(indices).reshape(shape).astype(labels.dtype)

    # Interpolate and return in image shape
    return image, labels


def elastic_transform_3d(image, labels, alpha, sigma, bg_val=0.0):
    """
    Elastic deformation of images as described in [Simard2003]_.
    [Simard2003] Simard, Steinkraus and Platt, "Best Practices for
    Convolutional Neural Networks applied to Visual Document Analysis", in
    Proc. of the International Conference on Document Analysis and
    Recognition, 2003.

    Modified from:
    https://gist.github.com/chsasank/4d8f68caf01f041a6453e67fb30f8f5a

    Modified to take 3 and 4 dimensional inputs
    Deforms both the image and corresponding label file
    image tri-linear interpolated
    Label volumes nearest neighbour interpolated
    """
    if image.ndim == 3:
        image = np.expand_dims(image, axis=-1)
    shape = image.shape[:3]
    channels = image.shape[-1]
    dtype = image.dtype
    bg_val = bg_val if isinstance(bg_val, (list, tuple, np.ndarray)) \
        else [bg_val] * channels

    # Define coordinate system
    coords = np.arange(shape[0]), np.arange(shape[1]), np.arange(shape[2])

    # Initialize interpolators
    im_intrps = []
    for i in range(channels):
        im_intrps.append(RegularGridInterpolator(coords, image[..., i],
                                                 method="linear",
                                                 bounds_error=False,
                                                 fill_value=bg_val[i],
                                                 dtype=np.float32))

    # Get random elastic deformations
    dx = gaussian_filter((np.random.rand(*shape) * 2 - 1), sigma,
                         mode="constant", cval=0.) * alpha
    dy = gaussian_filter((np.random.rand(*shape) * 2 - 1), sigma,
                         mode="constant", cval=0.) * alpha
    dz = gaussian_filter((np.random.rand(*shape) * 2 - 1), sigma,
                         mode="constant", cval=0.) * alpha

    # Define sample points
    x, y, z = np.mgrid[0:shape[0], 0:shape[1], 0:shape[2]]
    indices = np.reshape(x + dx, (-1, 1)), \
              np.reshape(y + dy, (-1, 1)), \
              np.reshape(z + dz, (-1, 1))

    # Interpolate all image channels
    image = np.empty(shape=image.shape, dtype=dtype)
    for i, intrp in enumerate(im_intrps):
        image[..., i] = intrp(indices).reshape(shape)

    # Interpolate labels
    if labels is not None:
        lab_intrp = RegularGridInterpolator(coords, labels,
                                            method="nearest",
                                            bounds_error=False,
                                            fill_value=0,
                                            dtype=np.uint8)

        labels = lab_intrp(indices).reshape(shape).astype(labels.dtype)

    # Interpolate and return in image shape
    return image, labels



from mpunet.logging import ScreenLogger
import os


def init_model(build_hparams, logger=None):
    from mpunet import models
    logger = logger or ScreenLogger()

    # Build new model of the specified type
    cls_name = build_hparams["model_class_name"]
    logger("Creating new model of type '%s'" % cls_name)

    return models.__dict__[cls_name](logger=logger, **build_hparams)


def model_initializer(hparams, continue_training, project_dir,
                      initialize_from=None, logger=None):
    logger = logger or ScreenLogger()

    # Init model
    model = init_model(hparams["build"], logger)

    if continue_training:
        if initialize_from:
            raise ValueError("Failed to initialize model with both "
                             "continue_training and initialize_from set.")
        from mpunet.utils import get_last_model, get_lr_at_epoch, \
                                          clear_csv_after_epoch, get_last_epoch
        model_path, epoch = get_last_model(os.path.join(project_dir, "model"))
        if model_path:
            model.load_weights(model_path, by_name=True)
            model_name = os.path.split(model_path)[-1]
        else:
            model_name = "<No model found>"
        csv_path = os.path.join(project_dir, "logs", "training.csv")
        if epoch == 0:
            epoch = get_last_epoch(csv_path)
        else:
            if epoch is None:
                epoch = 0
            clear_csv_after_epoch(epoch, csv_path)
        hparams["fit"]["init_epoch"] = epoch+1

        # Get the LR at the continued epoch
        lr, name = get_lr_at_epoch(epoch, os.path.join(project_dir, "logs"))
        if lr:
            hparams["fit"]["optimizer_kwargs"][name] = lr

        logger("[NOTICE] Training continues from:\n"
               "Model: %s\n"
               "Epoch: %i\n"
               "LR:    %s" % (model_name, epoch, lr))
    else:
        hparams["fit"]["init_epoch"] = 0
        if initialize_from:
            model.load_weights(initialize_from, by_name=True)
            logger("[NOTICE] Initializing parameters from:\n"
                   "{}".format(initialize_from))
    return model



import tensorflow as tf
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Layer
from tensorflow.keras.initializers import constant
from mpunet.logging import ScreenLogger
from mpunet.evaluate.loss_functions import SparseGeneralizedDiceLoss


def reg(W):
    r = tf.reduce_sum(tf.square(W)) / tf.cast(tf.size(W), tf.float32)
    return 1e-6 * r


class FusionLayer(Layer):
    def __init__(self, logger=None, activation_func=None, **kwargs):
        self.logger = logger or ScreenLogger()
        self.W, self.b = None, None
        self.activation = activation_func or tf.nn.softmax
        super(FusionLayer, self).__init__(**kwargs)

    def build(self, input_shape):

        self.W = self.add_weight(name='W',
                                 shape=(input_shape[1], input_shape[2]),
                                 initializer=constant(1.0),
                                 trainable=True,
                                 regularizer=reg)
        self.b = self.add_weight(name="b",
                                 shape=(1, input_shape[2]),
                                 initializer=constant(0.0),
                                 trainable=True,
                                 regularizer=reg)

        # Build the layer from the base class
        # This also sets self.built = True
        super(FusionLayer, self).build(input_shape)

    def call(self, x, *args, **kwargs):
        return self.activation(tf.reduce_sum(self.W * x, axis=1) + self.b)

    def compute_output_shape(self, input_shape):
        return None, input_shape[2]


class FusionModel(Model):
    def __init__(self, n_inputs, n_classes, weight="Simple", logger=None,
                 verbose=True):
        self.n_inputs = n_inputs
        self.n_classes = n_classes

        # Set Logger object
        self.logger = logger or ScreenLogger()

        # Set loss
        self.loss = SparseGeneralizedDiceLoss(tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE,
                                              type_weight=weight)

        # Init model
        super().__init__(*self.init_model(n_inputs, n_classes))

        if verbose:
            self._log()

    def init_model(self, n_inputs, n_classes):
        inputs = Input(shape=(n_inputs, n_classes))
        fusion = FusionLayer(activation_func=tf.nn.softmax)(inputs)

        return [inputs], [fusion]

    def _log(self):
        self.logger("Optimizer:  %s" % self.optimizer)
        self.logger("Loss:       %s" % self.loss)
        self.logger("Input:      %s" % self.input)
        self.logger("Output:     %s" % self.output)
        self.logger("N weights:  %s" % self.count_params())



from functools import wraps
from inspect import signature


def accepts(*types):
    def decorator(function):
        _types = types
        f_sig = signature(function)
        if "self" in f_sig.parameters:
            _types = (object,) + _types
        assert len(_types) == len(f_sig.parameters)

        @wraps(function)
        def wrapper(*args, **kwargs):
            for i, (o, t) in enumerate(zip(args, _types)):
                valid = False
                if isinstance(t, (tuple, list)):
                    for tt in t:
                        if isinstance(o, tt):
                            valid = True
                else:
                    if isinstance(o, t):
                        valid = True
                if not valid:
                    raise ValueError("Invalid input passed to '%s'. "
                                     "Expected type(s) %s (got %s)" % (function.__name__,
                                                                       t,
                                                                       type(o).__name__))
            return function(*args, **kwargs)
        return wrapper
    return decorator



from tensorflow.keras.callbacks import ModelCheckpoint
import warnings
import os


class ModelCheckPointClean(ModelCheckpoint):
    """
    Fixes bug in ModelCheckPoint that will fail to overwrite model/weight files
    of which the name changes as fitting progresses (epoch in name etc.)

    Overwrites the on_epoch_end method
    """
    def __init__(self, org_model=None, *args, **kwargs):
        ModelCheckpoint.__init__(self, *args, **kwargs)
        self.org_model = org_model
        self.last_file = None

    @property
    def __model(self):
        if self.org_model:
            return self.org_model
        else:
            return self.model

    def on_epoch_end(self, epoch, logs=None):
        print()
        logs = logs or {}
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0
            filepath = self.filepath.format(epoch=epoch, **logs)

            # Make root folder if not existing
            folder = os.path.split(os.path.abspath(filepath))[0]
            if not os.path.exists(folder):
                os.mkdir(folder)

            if self.save_best_only:
                current = logs.get(self.monitor)
                if current is None:
                    warnings.warn('Can save best model only with %s available, '
                                  'skipping.' % (self.monitor), RuntimeWarning)
                else:
                    if self.monitor_op(current, self.best):
                        if self.verbose > 0:
                            print('Epoch %05d: %s improved from %0.5f to %0.5f,'
                                  ' saving model to %s'
                                  % (epoch, self.monitor, self.best,
                                     current, filepath))
                        self.best = current
                        if self.last_file:
                            # Make sure we remove the file even with changing
                            # filename over fitting
                            os.remove(self.last_file)
                        self.last_file = filepath
                        if self.save_weights_only:
                            self.__model.save_weights(filepath, overwrite=True)
                        else:
                            self.__model.save(filepath, overwrite=True)
                    else:
                        if self.verbose > 0:
                            print('Epoch %05d: %s did not improve' %
                                  (epoch, self.monitor))
            else:
                if self.verbose > 0:
                    print('Epoch %05d: saving model to %s' % (epoch, filepath))
                if self.save_weights_only:
                    self.__model.save_weights(filepath, overwrite=True)
                else:
                    self.__model.save(filepath, overwrite=True)



import nibabel as nib
import numpy as np
import numpy.linalg as npl
import random
from mpunet.interpolation.linalg import mgrid_to_points, points_to_mgrid
from mpunet.interpolation.linalg import get_rotation_matrix


def get_pix_dim(nii_image):
    return nii_image.header["pixdim"][1:4]


def get_real_image_size(nii_image):
    pix_dims = get_pix_dim(nii_image)
    shape = np.asarray(nii_image.shape)[:3]
    return shape * pix_dims


def get_maximum_real_dim(nii_image):
    return np.max(get_real_image_size(nii_image))


def get_bounding_sphere_radius(nii_image):
    return np.linalg.norm(nii_image.center)


def get_bounding_sphere_real_radius(nii_image):
    real_dim = get_real_image_size(nii_image)
    return np.linalg.norm(real_dim / 2)


def get_maximum_real_dim_in_folder(folder):
    import os
    sizes = []
    for f in os.listdir(folder):
        if os.path.splitext(f)[-1] not in (".nii", ".gz"):
            continue
        else:
            f = os.path.join(folder, f)
        im = nib.load(f)
        sizes.append(get_maximum_real_dim(im))
    return np.max(sizes)


def get_voxel_grid(images, as_points=False):
    shape = images.shape[:3]
    grid = np.mgrid[0:shape[0]:1,
                    0:shape[1]:1,
                    0:shape[2]:1]

    if as_points:
        return mgrid_to_points(grid)
    else:
        return grid


def get_angle(v1, v2):
    v1_u = v1 / npl.norm(v1)
    v2_u = v2 / npl.norm(v2)
    return np.rad2deg(np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0)))


def get_voxel_axes_real_space(image, affine, return_basis=False):
    g_xx, g_yy, g_zz = get_voxel_axes(image)

    # Extract orthonormal basis matrix
    assert affine.shape == (4, 4)
    basis = affine[:-1, :-1]

    # Make sure basis is aligned with regular grid
    # Otherwise, rotate the grid and store transformation for later use
    pixdims = np.linalg.norm(basis, axis=0)
    transform = np.diag(pixdims)
    # sign = np.sign([np.dot(transform[:, i], basis[:, i]) for i in range(3)])
    # transform = np.diag(sign).dot(transform)

    if np.any(~np.isclose(transform, basis)):
        rot_mat = transform.dot(np.linalg.inv(basis))
    else:
        rot_mat = None

    # Get grid in real space
    g_xx = g_xx * transform[0, 0]
    g_yy = g_yy * transform[1, 1]
    g_zz = g_zz * transform[2, 2]

    if return_basis:
        return (g_xx, g_yy, g_zz), transform, rot_mat
    else:
        return g_xx, g_yy, g_zz


def get_voxel_axes(image):
    x, y, z, _ = image.shape
    g_xx = np.arange(x, dtype=np.float32) - (x - 1) / 2
    g_yy = np.arange(y, dtype=np.float32) - (y - 1) / 2
    g_zz = np.arange(z, dtype=np.float32) - (z - 1) / 2
    return g_xx, g_yy, g_zz


def get_voxel_grid_real_space(images, append_ones=False):
    # Get shape excluding channels
    shape = images.shape[:-1]

    # Get affine transforming voxel positions to real space positions
    vox_to_real_affine = images.affine[:-1, :-1]

    # Transform axes from voxel space to real space
    grid_vox_space = np.mgrid[0:shape[0]:1,
                              0:shape[1]:1,
                              0:shape[2]:1]

    # Move grid to real space
    grid_points_real_space = vox_to_real_affine.dot(
        mgrid_to_points(grid_vox_space).T).T

    # Center
    centered_grid_points_real_space = grid_points_real_space - \
                                      np.mean(grid_points_real_space, axis=0)

    # Append column of ones?
    if append_ones:
        centered_grid_points_real_space = np.column_stack(
            (grid_points_real_space,
             np.ones(len(grid_points_real_space))))

    # Return real space grid as mgrid
    points = points_to_mgrid(centered_grid_points_real_space, shape)

    return points


def get_random_views(N, dim=3, norm=np.random.normal, pos_z=True, weights=None):
    """
    http://en.wikipedia.org/wiki/N-sphere#Generating_random_points
    """
    normal_deviates = norm(size=(N, dim))
    radius = np.linalg.norm(normal_deviates, axis=1)[:, np.newaxis]
    views = normal_deviates / radius
    if pos_z:
        views[:, -1] = np.abs(views[:, -1])

    if weights is not None:
        v_weghted = views * weights
        views = v_weghted / np.linalg.norm(v_weghted, axis=1)[:, np.newaxis]

    return views


def sample_random_views_with_angle_restriction(views, min_angle_deg,
                                               auditor=None, logger=None):
    from itertools import combinations
    from mpunet.logging import ScreenLogger
    logger = logger or ScreenLogger()
    logger("Generating %i random views..." % views)

    # Weight by median sample resolution along each axis
    if auditor is not None:
        res = np.median(auditor.info["pixdims"], axis=0)
        logger("[OBS] Weighting random views by median res: %s" % res)
    else:
        res = None

    N = views
    found = False
    tries = 0
    while not found:
        tries += 1
        views = get_random_views(N, dim=3, pos_z=True, weights=res)
        angles = [get_angle(v1, v2) for v1, v2 in combinations(views, 2)]
        found = np.all(np.asarray(angles) > min_angle_deg)
        min_angle_deg -= 1
    return views


def sample_plane(norm_vector, sample_dim, real_space_span,
                 real_space_sample_sphere_radius, noise_sd=0.,
                 return_real_space_grid=False):
    # Sample a random displacement
    # Get random displacement within sample sphere
    rd = np.random.randint(-real_space_sample_sphere_radius,
                           real_space_sample_sphere_radius, 1)[0]

    return sample_plane_at(norm_vector=norm_vector,
                           sample_dim=sample_dim,
                           real_space_span=real_space_span,
                           offset_from_center=rd,
                           noise_sd=noise_sd,
                           test_mode=return_real_space_grid)


def sample_plane_at(norm_vector, sample_dim, real_space_span,
                    offset_from_center, noise_sd, test_mode=False):
    # Prepare normal vector to the plane
    n_hat = np.array(norm_vector, np.float32)
    n_hat /= np.linalg.norm(n_hat)

    # Add noise?
    if type(noise_sd) is not np.ndarray:
        noise_sd = np.random.normal(scale=noise_sd, size=3)

    n_hat += noise_sd
    n_hat /= np.linalg.norm(n_hat)

    if np.all(n_hat[:-1] < 0.2):
        # Vector pointing primarily up, noise will have large effect on image
        # orientation. We force the first two components to go into the
        # positive direction to control variability of sampling
        n_hat[:-1] = np.abs(n_hat[:-1])
    if np.all(np.isclose(n_hat[:-1], 0)):
        u = np.array([1, 0, 0])
        v = np.array([0, 1, 0])
    else:
        # Find vector in same vertical plane as nhat
        nhat_vs = n_hat.copy()
        nhat_vs[-1] = nhat_vs[-1] + 1
        nhat_vs /= np.linalg.norm(nhat_vs)

        # Get two orthogonal vectors in plane, u pointing down in z-direction
        u = get_rotation_matrix(np.cross(n_hat, nhat_vs), -90).dot(n_hat)
        v = np.cross(n_hat, u)

    # Define basis matrix + displacement to center (affine transformation)
    basis = np.column_stack((u, v, n_hat))

    # Define regular grid (centered at origin)
    hd = real_space_span // 2
    g = np.linspace(-hd, hd, sample_dim)

    j = complex(sample_dim)
    grid = np.mgrid[-hd:hd:j,
                    -hd:hd:j,
                    offset_from_center:offset_from_center:1j]

    # Calculate voxel coordinates on the real space grid
    points = mgrid_to_points(grid)

    real_points = basis.dot(points.T).T
    real_grid = points_to_mgrid(real_points, grid.shape[1:])

    if test_mode:
        return real_grid, g, np.linalg.inv(basis)
    else:
        return real_grid


def sample_box(sample_dim, real_box_dim, real_dims, noise_sd=0., test_mode=False):

    # Set sample space equal to real_dims or expanded to 1.1x sample box dim
    # 1.1x to give a little room around the image for sampling
    sample_space = np.asarray([max(i, real_box_dim*1.1) for i in real_dims])

    # Sample a random displacement
    # Get random displacement within sample space, center on origin
    d = (sample_space - real_box_dim)
    placement = np.array([random.uniform(0, d[i]) for i in range(3)]) - sample_space/2

    return sample_box_at(real_placement=placement,
                         sample_dim=sample_dim,
                         real_box_dim=real_box_dim,
                         noise_sd=noise_sd,
                         test_mode=test_mode)


def sample_box_at(real_placement, sample_dim, real_box_dim,
                  noise_sd, test_mode):

    j = complex(sample_dim)
    a, b, c = real_placement
    grid = np.mgrid[a:a + real_box_dim:j,
                    b:b + real_box_dim:j,
                    c:c + real_box_dim:j]

    rot_mat = np.eye(3)
    rot_grid = grid
    if noise_sd:
        # Get random rotation vector
        rot_axis = get_random_views(N=1, dim=3, pos_z=True)

        rot_angle = False
        while not rot_angle:
            angle = np.abs(np.random.normal(scale=noise_sd, size=1)[0])
            if angle < 2*np.pi:
                rot_angle = angle

        rot_mat = get_rotation_matrix(rot_axis, angle_rad=rot_angle)

        # Center --> apply rotation --> revert centering --> mgrid
        points = mgrid_to_points(grid)
        center = np.mean(points, axis=0)
        points -= center
        points = rot_mat.dot(points.T).T + center
        rot_grid = points_to_mgrid(points, grid.shape[1:])

    if test_mode:
        axes = (np.linspace(a, a+real_box_dim, sample_dim),
                np.linspace(b, b+real_box_dim, sample_dim),
                np.linspace(c, c+real_box_dim, sample_dim))
        return rot_grid, axes, np.linalg.inv(rot_mat)
    else:
        return rot_grid



from .elastic_deformation import elastic_transform_2d, elastic_transform_3d
import numpy as np


class Augmenter(object):
    """
    Not yet implemented
    """
    def __call__(self, batch_x, batch_y, bg_values, batch_w=None):
        raise NotImplemented


class Elastic(Augmenter):
    """
    2D and 3D random elastic deformations augmenter base class

    Applies either Elastic2D or Elastic3D to every element of a batch of images
    """
    def __init__(self, alpha, sigma, apply_prob,
                 transformer_func, aug_weight=0.33):
        """
        Args:
            alpha: A number of tuple/list of two numbers specifying a range
                   of alpha values to sample from in each augmentation call
                   The alpha value determines the strength of the deformation
            sigma: A number of tuple/list of two numbers specifying a range
                   of sigma values to sample from in each augmentation call
                   The sigma value determines the smoothness of the deformation
            apply_prob: Apply the transformation only with some probability
                        Otherwise, return the image untransformed
            transformer_func: The deformation function, either Elastic2D or
                              Elastic3D
            aug_weight: If a list of weights of len(batch_x) elements is passed
                        the aug_weight will replace the passed weight at index
                        i if image i in batch_x is transformed.
                        This allows for assigning a different weight to images
                        that were augmented versus real images.
        """
        # Initialize base
        super().__init__()

        if isinstance(alpha, (list, tuple)):
            if len(alpha) != 2:
                raise ValueError("Invalid list of alphas specified '%s'. "
                                 "Should be 2 numbers." % alpha)
            if alpha[1] <= alpha[0]:
                raise ValueError("alpha upper is smaller than sigma lower (%s)" % alpha)
        if isinstance(sigma, (list, tuple)):
            if len(sigma) != 2:
                raise ValueError("Invalid list of sigmas specified '%s'. "
                                 "Should be 2 numbers." % sigma)
            if sigma[1] <= sigma[0]:
                raise ValueError("Sigma upper is smaller than sigma lower (%s)" % sigma)
        if apply_prob > 1 or apply_prob < 0:
            raise ValueError("Apply probability is invalid with value %3.f" % apply_prob)

        self._alpha = alpha
        self._sigma = sigma
        self.apply_prob = apply_prob
        self.trans_func = transformer_func
        self.weight = aug_weight
        self.__name__ = "Elastic"

    @property
    def alpha(self):
        """
        Return a randomly sampled alpha value in the range [alpha[0], alpha[1]]
        or return the integer/float alpha if alpha is not a list
        """
        if isinstance(self._alpha, (list, tuple)):
            return np.random.uniform(self._alpha[0], self._alpha[1], 1)[0]
        else:
            return self._alpha

    @property
    def sigma(self):
        """
        Return a randomly sampled sigma value in the range [sigma[0], sigma[1]]
        or return the integer/float sigma if sigma is not a list
        """
        if isinstance(self._sigma, (list, tuple)):
            return np.random.uniform(self._sigma[0], self._sigma[1], 1)[0]
        else:
            return self._sigma

    def __call__(self, batch_x, batch_y, bg_values, batch_w=None):
        """
        Deform all images in a batch of images (using linear intrp) and
        corresponding labels (using nearest intrp)
        """
        # Only augment some of the images (determined by apply_prob)
        augment_mask = np.random.rand(len(batch_x)) <= self.apply_prob

        augmented_x, augmented_y = [], []
        for i, (augment, x, y, bg_vals) in enumerate(zip(augment_mask,
                                                         batch_x, batch_y,
                                                         bg_values)):
            if augment:
                x, y = self.trans_func(x, y, self.alpha, self.sigma, bg_vals)
                if batch_w is not None:
                    batch_w[i] = self.weight
            augmented_x.append(x)
            augmented_y.append(y)

        if batch_w is not None:
            return augmented_x, augmented_y, batch_w
        else:
            return augmented_x, augmented_y

    def __str__(self):
        return "%s(alpha=%s, sigma=%s, apply_prob=%.3f)" % (
            self.__name__, self._alpha, self._sigma, self.apply_prob
        )

    def __repr__(self):
        return str(self)


class Elastic2D(Elastic):
    """
    Applies random elastic deformations to 2D images.

    See docstring of Elastic (base class)
    """
    def __init__(self, alpha, sigma, apply_prob):
        """
        See docstring of Elastic (base class)
        """
        super().__init__(alpha, sigma, apply_prob,
                         transformer_func=elastic_transform_2d)
        self.__name__ = "Elastic2D"


class Elastic3D(Elastic):
    """
    Applies random elastic deformations to 3D images.

    See docstring of Elastic (base class)
    """
    def __init__(self, alpha, sigma, apply_prob):
        """
        See docstring of Elastic (base class)
        """
        super().__init__(alpha, sigma, apply_prob,
                         transformer_func=elastic_transform_3d)
        self.__name__ = "Elastic3D"

    def __str__(self):
        return "Elastic3D(alpha=%s, sigma=%s, apply_prob=%.3f)" % (
            self._alpha, self._sigma, self.apply_prob
        )



import re
import os
import numpy as np
import glob
import contextlib


def _get_system_wide_set_gpus():
    allowed_gpus = os.environ.get("CUDA_VISIBLE_DEVICES")
    if allowed_gpus:
        allowed_gpus = allowed_gpus.replace(" ", "").split(",")
    return allowed_gpus


def get_free_gpus(max_allowed_mem_usage=400):
    # Check if allowed GPUs are set in CUDA_VIS_DEV.
    allowed_gpus = _get_system_wide_set_gpus()
    if allowed_gpus:
        print("[OBS] Considering only system-wise allowed GPUs: {} (set in"
              " CUDA_VISIBLE_DEVICES env variable).".format(allowed_gpus))
        return allowed_gpus
    # Else, check GPUs on the system and assume all non-used (mem. use less
    # than max_allowed_mem_usage) is fair game.
    from subprocess import check_output
    try:
        # Get list of GPUs
        gpu_list = check_output(["nvidia-smi", "-L"], universal_newlines=True)
        gpu_ids = np.array(re.findall(r"GPU[ ]+(\d+)", gpu_list), dtype=np.int)

        # Query memory usage stats from nvidia-smi
        output = check_output(["nvidia-smi", "-q", "-d", "MEMORY"],
                              universal_newlines=True)

        # Fetch the memory usage of each GPU
        mem_usage = re.findall(r"FB Memory Usage.*?Used[ ]+:[ ]+(\d+)",
                               output, flags=re.DOTALL)
        assert len(gpu_ids) == len(mem_usage)

        # Return all GPU ids for which the memory usage is exactly 0
        free = list(map(lambda x: int(x) <= max_allowed_mem_usage, mem_usage))
        return list(gpu_ids[free])
    except FileNotFoundError as e:
        raise FileNotFoundError("[ERROR] nvidia-smi is not installed. "
                                "Consider setting the --num_GPUs=0 flag.") from e


def _get_free_gpu(free_GPUs, N=1):
    try:
        free_gpu = ",".join(map(str, free_GPUs[0:N]))
    except IndexError as e:
        raise OSError("No GPU available.") from e
    return free_gpu


def get_free_gpu(N=1):
    free = get_free_gpus()
    return _get_free_gpu(free, N=N)


def await_and_set_free_gpu(N=1, sleep_seconds=60, logger=None):
    gpu = ""
    if N != 0:
        from time import sleep
        logger = logger or print
        logger("Waiting for free GPU.")
        found_gpu = False
        while not found_gpu:
            gpu = get_free_gpu(N=N)
            if gpu:
                logger("Found free GPU: %s" % gpu)
                found_gpu = True
            else:
                logger("No available GPUs... Sleeping %i seconds." % sleep_seconds)
                sleep(sleep_seconds)
    set_gpu(gpu)


def set_gpu(gpu):
    os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu)


def add_noise_to_views(views, sd):
    # Add Gaussian noise to views specified in parameter file
    return [np.array(v, dtype=np.float32) + np.random.normal(scale=sd, size=3)
            for v in views]


def get_best_model(model_dir):
    if len(os.listdir(model_dir)) == 0:
        raise OSError("Model dir {} is empty.".format(model_dir))
    # look for models, order: val_dice, val_loss, dice, loss, model_weights
    patterns = [
        ("@epoch*val_dice*", np.argmax),
        ("@epoch*val_loss*", np.argmin),
        ("@epoch*dice*", np.argmax),
        ("@epoch*loss*", np.argmin)
    ]
    for pattern, select_func in patterns:
        models = glob.glob(os.path.join(model_dir, pattern))
        if models:
            scores = []
            for m in models:
                scores.append(float(re.findall(r"(\d+[.]\d+)", m)[0]))
            return os.path.abspath(models[select_func(np.array(scores))])
    m = os.path.abspath(os.path.join(model_dir, "model_weights.h5"))
    if not os.path.exists(m):
        raise OSError("Did not find any model files matching the patterns {} "
                      "and did not find a model_weights.h5 file."
                      "".format(patterns))
    return m


def get_last_model(model_dir):
    models = glob.glob(os.path.join(model_dir, "@epoch*"))
    epochs = []
    for m in models:
        epochs.append(int(re.findall(r"@epoch_(\d+)_", m)[0]))
    if epochs:
        last = np.argmax(epochs)
        return os.path.abspath(models[last]), int(epochs[int(last)])
    else:
        generic_path = os.path.join(model_dir, "model_weights.h5")
        if os.path.exists(generic_path):
            # Return epoch 0 as we dont know where else to start
            # This may be changed elsewhere in the code based on the
            # training data CSV file
            return generic_path, 0
        else:
            # Start from scratch, or handle as see fit at call point
            return None, None


def get_lr_at_epoch(epoch, log_dir):
    log_path = os.path.join(log_dir, "training.csv")
    if not os.path.exists(log_path):
        print("No training.csv file found at %s. Continuing with default "
              "learning rate found in parameter file." % log_dir)
        return None, None
    import pandas as pd
    df = pd.read_csv(log_path)
    possible_names = ("lr", "LR", "learning_rate", "LearningRate")
    try:
        in_df = [l in df.columns for l in possible_names].index(True)
    except ValueError:
        return None, None
    col_name = possible_names[in_df]
    return float(df[col_name][int(epoch)]), col_name


def clear_csv_after_epoch(epoch, csv_file):
    if os.path.exists(csv_file):
        import pandas as pd
        try:
            df = pd.read_csv(csv_file)
        except pd.errors.EmptyDataError:
            # Remove the file
            os.remove(csv_file)
            return
        # Remove any trailing runs and remove after 'epoch'
        try:
            df = df[np.flatnonzero(df["epoch"] == 0)[-1]:]
        except IndexError:
            pass
        df = df[:epoch+1]
        # Save again
        with open(csv_file, "w") as out_f:
            out_f.write(df.to_csv(index=False))


def get_last_epoch(csv_file):
    epoch = 0
    if os.path.exists(csv_file):
        import pandas as pd
        df = pd.read_csv(csv_file)
        epoch = int(df["epoch"].to_numpy()[-1])
    return epoch


def set_bias_weights_on_all_outputs(model, data_queue, hparams, logger):
    # This will bias the softmax output layer to output class confidences
    # equal to the class frequency
    if hasattr(model, "out_layers"):
        raise NotImplementedError("Multi task models not yet supported in "
                                  "mpunet >= 0.2.6")
        # Multiple output layers, set bias for each
        layers = model.out_layers
        loaders = [t.image_pair_loader for t in data_queue]
    else:
        layers = [None]
        for layer in model.layers[::-1]:
            # Start from last layer and go up until one that has an activation
            # funtion is met (note: this skips layers like Reshape, Cropping
            # etc.)
            if hasattr(layer, 'activation'):
                layers = [layer]
                break
        data_queues = [data_queue]
    for layer, data_queue in zip(layers, data_queues):
        set_bias_weights(layer=layer,
                         data_queue=data_queue,
                         class_counts=hparams.get("class_counts"),
                         logger=logger)


def set_bias_weights(layer, data_queue, class_counts=None, logger=None):
    if layer.activation.__name__ != "softmax":
        raise ValueError("Setting output layer bias currently only supported "
                         "with softmax activation functions. Output layer has "
                         "'%s'" % layer.activation.__name__)
    from mpunet.logging.default_logger import ScreenLogger
    logger = logger or ScreenLogger()
    # Get original and set new weights
    weights = layer.get_weights()
    if len(weights) != 2:
        raise ValueError("Output layer does not have bias weights.")
    bias_shape = weights[-1].shape
    n_classes = weights[-1].size

    # Estimate counts if not specified
    if class_counts is None:
        class_counts = np.zeros(shape=[n_classes], dtype=np.int)
        if hasattr(data_queue, 'max_loaded'):
            # Limitation queue, count once for each image currently in queue
            n_images = data_queue.max_loaded
        else:
            n_images = len(data_queue.dataset)
        logger("OBS: Estimating class counts from {} images".format(n_images))
        for _ in range(n_images):
            with data_queue.get_random_image() as image:
                class_counts += np.bincount(image.labels.ravel(),
                                            minlength=n_classes)

    # Compute frequencies
    freq = np.asarray(class_counts/np.sum(class_counts))

    # Compute bias weights
    bias = np.log(freq * np.sum(np.exp(freq)))
    bias /= np.linalg.norm(bias)
    weights[-1] = bias.reshape(bias_shape)

    layer.set_weights(weights)
    logger("Setting bias weights on output layer to:\n%s" % bias)


def get_confidence_dict(path, views):
    # Calculate linearly distributed per class weights between the models
    conf_paths = [path + "/validation_confidence_%s.npz" % v for v in views]
    confs = np.column_stack([np.load(x)["arr_0"] for x in conf_paths])
    confs = (confs/np.sum(confs, axis=1).reshape((confs.shape[0], 1))).T

    # Add each confidence array to a dictionary under key <view>
    confs = {str(v): confs[i] for i, v in enumerate(views)}

    return confs


def random_split(X, y, fraction):
    # Take random split of validation data
    n_val = int(X.shape[0] * fraction)
    val_ind = np.random.choice(np.arange(X.shape[0]), size=n_val)
    X_val, y_val = X[val_ind], y[val_ind]

    # Get inverse for training set
    X, y = np.delete(X, val_ind, axis=0), np.delete(y, val_ind, axis=0)

    return X, y, X_val, y_val


def create_folders(folders, create_deep=False):
    def safe_make(path, make_func):
        try:
            make_func(path)
        except FileExistsError:
            # If running many jobs in parallel this may occur
            pass
    make_func = os.mkdir if not create_deep else os.makedirs
    if isinstance(folders, str):
        if not os.path.exists(folders):
            safe_make(folders, make_func)
    else:
        folders = list(folders)
        for f in folders:
            if f is None:
                continue
            if not os.path.exists(f):
                safe_make(f, make_func)


@contextlib.contextmanager
def print_options_context(*args, **kwargs):
    original = np.get_printoptions()
    np.set_printoptions(*args, **kwargs)
    try:
        yield
    finally:
        np.set_printoptions(**original)


def arr_to_fixed_precision_string(arr, precision):
    f = np.format_float_positional
    s = map(lambda x: f(x, precision, pad_right=precision), arr)
    return "[{}]".format(" ".join(s))


class DummyContext(object):
    def __enter__(self): return self

    def __exit__(*x): pass


def pred_to_class(tensor, img_dims=3, threshold=0.5, has_batch_dim=False):
    tensor_dim = img_dims + int(has_batch_dim)
    dims = len(tensor.shape)
    if dims == tensor_dim:
        # Check if already integer targets
        if np.issubdtype(tensor.dtype, np.integer):
            return tensor
        else:
            return tensor >= threshold
    elif tensor.shape[-1] == 1:
        if np.issubdtype(tensor.dtype, np.integer):
            # Squeeze last axis
            return np.squeeze(tensor)
        else:
            return tensor >= threshold
    else:
        # Convert predicted probabilities to predicted class
        return tensor.argmax(-1).astype(np.uint8)


def highlighted(string):
    length = len(string) if "\n" not in string else max([len(s) for s in string.split("\n")])
    border = "-" * length
    return "%s\n%s\n%s" % (border, string, border)


def await_PIDs(PIDs, check_every=120):
    if isinstance(PIDs, str):
        for pid in PIDs.split(","):
            wait_for(int(pid), check_every=check_every)
    else:
        wait_for(PIDs, check_every=check_every)


def wait_for(PID, check_every=120):
    """
    Check for a running process with PID 'PID' and only return when the process
    is no longer running. Checks the process list every 'check_every' seconds.
    """
    if not PID:
        return
    if not isinstance(PID, int):
        try:
            PID = int(PID)
        except ValueError as e:
            raise ValueError("Cannot wait for PID '%s', must be an integer"
                             % PID) from e
    _wait_for(PID, check_every)


def _wait_for(PID, check_every=120):
    still_running = True
    import subprocess
    import time
    print("\n[*] Waiting for process PID=%i to terminate..." % PID)
    while still_running:
        ps = subprocess.Popen(('ps', '-p', "%i" % PID), stdout=subprocess.PIPE)
        try:
            output = subprocess.check_output(('grep', '%i' % PID), stdin=ps.stdout)
        except subprocess.CalledProcessError:
            output = False
        ps.wait()

        still_running = bool(output)
        if still_running:
            print("Process %i still running... (sleeping %i seconds)" % (PID, check_every))
            time.sleep(check_every)


def check_kwargs(kwargs, allowed, func=None):
    s = ("Function '{}': ".format(func.__name__)) if func is not None else ""
    for param in kwargs:
        if param not in allowed:
            raise RuntimeError("{}Unexpected parameter '{}' passed.".
                               format(s, param))


def ensure_list_or_tuple(obj):
    return [obj] if not isinstance(obj, (list, tuple)) else obj



"""
Mathias Perslev
MSc Bioinformatics

University of Copenhagen
November 2017
"""

import numpy as np
import nibabel as nib
from contextlib import contextmanager

from pathlib import Path
from mpunet.preprocessing import get_scaler
from mpunet.logging import ScreenLogger
from mpunet.interpolation.sample_grid import get_real_image_size, get_pix_dim
from mpunet.interpolation.view_interpolator import ViewInterpolator
from mpunet.utils import ensure_list_or_tuple

# Errors
from mpunet.errors.image_errors import ReadOnlyAttributeError

# w2 negative threshold is too strict for this data set
nib.Nifti1Header.quaternion_threshold = -1e-6


class ImagePair(object):
    """
    ImagePair
    ---
    Represents one data point of 1 .nii image file and corresponding labels
    """
    def __init__(self, img_path, labels_path=None, sample_weight=1.0,
                 bg_class=0, logger=None, im_dtype=np.float32,
                 lab_dtype=np.uint8):
        """
        Initializes the ImagePair object from two paths to .nii file images

        The following actions are taken immediately on initialization:
            - Image and label Nibabel objects are created and metadata is
              loaded, but image data is not yet loaded into memory
              Loading of data happens automatically at the first reference to
              the image or label attributes
            - In identifier name is established (filename minus extension)
            - Placeholder attributes are created that may be used in methods

        Args:
            img_path:      Path to a Nifti image file (.nii, .nii.gz)
            labels_path:   Path to a Nifti image file (.nii, .nii.gz)
                           Filename must be identical to the img_path filename!
                           Note: A labels_path can be omitted if no ground
                           truth label map exists
            sample_weight: A float value assigning an overall weight to the
                           image/label pair - used by some optimization schemas
            bg_class:       The background class integer value, usually 0
            logger:        A mpunet logger object writing to screen and
                           /or a logfile on disk
            im_dtype:      A numpy data type that the image will be cast to
            lab_dtype:     A numpy data type that the labels will be cast to
        """
        # Labels included?
        self.predict_mode = not labels_path

        # Set a weight for all slices fetches from this image
        # We set value trough property setter to validate input
        self._sample_weight = None
        self.sample_weight = sample_weight

        # Set logger or default print
        self.logger = logger if logger is not None else ScreenLogger()

        # Set image and label paths (absolute)
        self.image_path = self._validate_path(img_path)
        if not self.predict_mode:
            self.labels_path = self._validate_path(labels_path)

        # Validate that the image and label data match and get image identifier
        self.identifier = self._get_and_validate_id()

        # Set variables to store loaded image and label information
        self.image_obj = nib.load(self.image_path)
        self.labels_obj = None
        if not self.predict_mode:
            self.labels_obj = nib.load(self.labels_path)

        # Stores the data of the image and labels objects
        self._image = None
        self._labels = None
        self._scaler = None
        self._bg_value = None
        self._bg_class = int(bg_class)

        # ViewInterpolator object initialized with set_interpolator_object
        self._interpolator = None

        # Data types
        self.im_dtype = im_dtype
        self.lab_dtype = lab_dtype

    def __str__(self):
        return "ImagePair(id={}, shape={}, real_shape={}, loaded={})".format(
            self.identifier, self.shape, self.real_center, self.is_loaded
        )

    def __repr__(self):
        return self.__str__()

    def log_image(self, print_calling_method=False):
        """
        Log basic stats for this ImagePair.
        """
        self.logger("%s\n"
                    "--- loaded:     %s\n"
                    "--- shape:      %s\n"
                    "--- bg class    %i\n"
                    "--- bg value    %s\n"
                    '--- scaler      %s\n'
                    "--- real shape: %s\n"
                    "--- pixdim:     %s" % (
                        self.identifier,
                        self.is_loaded,
                        self.shape,
                        self._bg_class,
                        self._bg_value,
                        ensure_list_or_tuple(self._scaler)[0],
                        np.round(get_real_image_size(self), 3),
                        np.round(get_pix_dim(self), 3)
                    ), print_calling_method=print_calling_method)

    def _get_and_validate_id(self):
        """
        Validates if the image identifier and label identifier match.
        Returns the image identifier.
        """
        img_id = self.image_path.stem.split('.')[0]
        if not self.predict_mode:
            labels_id = self.labels_path.stem.split('.')[0]
            if img_id != labels_id:
                raise ValueError("Image identifier '%s' does not match labels identifier '%s'"
                                 % (img_id, labels_id))
        return img_id

    @property
    def affine(self):
        return self.image_obj.affine

    @affine.setter
    def affine(self, _):
        raise ReadOnlyAttributeError("Manually setting the affine attribute "
                                     "is not allowed. Initialize a new "
                                     "ImagePair object.")

    @property
    def header(self):
        return self.image_obj.header

    @header.setter
    def header(self, _):
        raise ReadOnlyAttributeError("Manually setting the header attribute "
                                     "is not allowed. Initialize a new "
                                     "ImagePair object.")

    @property
    def image(self):
        """
        Ensures image is loaded and then returns it
        Note that we load the Nibabel data with the caching='unchanged'. This
        means that the Nibabel Nifti1Image object does NOT maintain its own
        internal copy of the image. Un-assigning self._image will GC the array.
        """
        if self._image is None:
            self._image = self.image_obj.get_fdata(caching='unchanged',
                                                   dtype=self.im_dtype)
        if self._image.ndim == 3:
            self._image = np.expand_dims(self._image, -1)
        return self._image

    @image.setter
    def image(self, _):
        raise ReadOnlyAttributeError("Manually setting the image attribute "
                                     "is not allowed. Initialize a new "
                                     "ImagePair object.")

    @property
    def labels(self):
        """ Like self.image """
        if self._labels is None:
            try:
                self._labels = self.labels_obj.get_fdata(caching="unchanged").astype(self.lab_dtype)
            except AttributeError:
                return None
        return self._labels

    @labels.setter
    def labels(self, _):
        raise ReadOnlyAttributeError("Manually setting the labels "
                                     "attribute is not allowed. "
                                     "Initialize a new ImagePair object.")

    @staticmethod
    def _validate_path(path):
        path = Path(path)
        if path.exists() and path.suffix in (".nii", ".mat", ".gz"):
            return path
        else:
            raise FileNotFoundError("File '%s' not found or not a "
                                    ".nii or .mat file." % path)

    @property
    def estimated_memory(self):
        """
        Note this will overestimate the memory footprint, actual memory usage
        will never be above this estimation.
        """
        import pickle
        return len(pickle.dumps(self))

    @property
    def sample_weight(self):
        return self._sample_weight

    @sample_weight.setter
    def sample_weight(self, weight):
        try:
            weight = float(weight)
        except ValueError:
            raise ValueError("Sample weight must be a numeric type (got '%s')"
                             % type(weight))
        if weight <= 0 or weight > 1:
            raise ValueError("Sample weight must be greater than "
                             "0 and less than or equal to 1")
        self._sample_weight = weight

    @property
    def center(self):
        """
        Returns:
            The voxel-space center of the image
        """
        return (self.shape[:-1]-1)/2

    @property
    def real_center(self):
        """
        Returns:
            The scanner-space center of the image
        """
        return self.affine[:3, :3].dot(self.center) + self.affine[:3, -1]

    @property
    def shape(self):
        """
        Returns:
            The voxel shape of the image (always rank 4 with channels axis)
        """
        s = np.asarray(self.image_obj.shape)
        if len(s) == 3:
            s = np.append(s, 1)
        return s

    @property
    def real_shape(self):
        """
        Returns:
            The real (physical, scanner-space span) shape of the image
        """
        return get_real_image_size(self.image_obj)

    @property
    def n_channels(self):
        return self.shape[-1]

    @property
    def bg_class(self):
        return self._bg_class

    @bg_class.setter
    def bg_class(self, _):
        raise ReadOnlyAttributeError("Cannot set a new background class. "
                                     "Initialize a new ImagePair object.")

    @property
    def bg_value(self):
        if self._bg_value is None or isinstance(self._bg_value[0], str):
            self.set_bg_value(self._bg_value, compute_now=True)
        return self._bg_value

    @bg_value.setter
    def bg_value(self, _):
        raise ReadOnlyAttributeError("New background values must be set in the"
                                     " self.set_bg_value method.")

    def set_bg_value(self, bg_value, compute_now=False):
        """
        Set a new background value for this ImagePair.

        Args:
            bg_value:    A value defining the space outside of the image region
                         bg_value may be a number of string of the format
                         '[0-100]'pct specifying a percentile value to compute
                         across the image and use for bg_value.
            compute_now: If a percentile string was passed, compute the
                         percentile now (True) or lazily when accessed via
                         self.bg_value at a later time (False).
        """
        bg_value = self.standardize_bg_val(bg_value)
        if compute_now and isinstance(bg_value[0], str):
            bg_value = self._bg_pct_string_to_value(bg_value)
        self._bg_value = bg_value

    @property
    def scaler(self):
        """ Return the currently set scaler object """
        if isinstance(self._scaler, tuple):
            self.set_scaler(*self._scaler, compute_now=True)
        return self._scaler

    @scaler.setter
    def scaler(self, _):
        raise ReadOnlyAttributeError("New scalers must be set with the "
                                     "self.set_scaler method.")

    def set_scaler(self, scaler, ignore_less_eq=None, compute_now=False):
        """
        Sets a scaler on the ImagePair fit to the stored image
        See mpunet.preprocessing.scaling

        Args:
            scaler:         A string naming a sklearn scaler type
            ignore_less_eq: A float or list of floats. Only consider values
                            above 'ignore_less_eq' in a channel when scaling.
            compute_now:    Initialize (and load data if not already) the
                            scaler now. Otherwise, the scaler will be
                            initialized at access time.
        """
        if compute_now:
            scaler = get_scaler(scaler=scaler,
                                ignore_less_eq=ignore_less_eq).fit(self.image)
            self._scaler = scaler
        else:
            self._scaler = (scaler, ignore_less_eq)

    def apply_scaler(self):
        """
        Apply the stored scaler (channel-wise) to the stored image
        Note: in-place opperation
        """
        self._image = self.scaler.transform(self.image)

    @property
    def interpolator(self):
        """
        Return a interpolator for this object.
        If not already set, will initialize an interpolator and store it
        """
        if not self._interpolator:
            self.set_interpolator_with_current()
        return self._interpolator

    @interpolator.setter
    def interpolator(self, _):
        raise ReadOnlyAttributeError("Cannot set the interpolator property. "
                                     "Call self.set_interpolator_with_current"
                                     " to set a new interpolator object.")

    @property
    def is_loaded(self):
        return self._image is not None

    def load(self):
        """
        Forces a load on any set attributes in
            (self.image, self.labels,
            self.interpolator, self.bg_value, self.scaler)
        """
        # OBS: Invoking the getter method on these properties invokes a load
        # No need to store the results!
        _ = [_ for _ in (self.image, self.labels, self.bg_value,
                         self.scaler, self.interpolator)]

    @contextmanager
    def loaded_in_context(self):
        """
        Context manager which keeps this ImagePair loaded in the context
        and unloads it at exit.
        """
        try:
            yield self.load()
        finally:
            self.unload()

    def unload(self, unload_scaler=False):
        """
        Unloads the ImagePair by un-assigning the image and labels attributes
        Also clears the currently set interpolator object, as this references
        the image and label arrays and thus prevents GC.

        Args:
            unload_scaler: boolean indicating whether or not to also clear the
                           currently set scaler object. Not clearing this may
                           be useful if the ImagePair is loaded/unloaded often
                           (in a data queue for instance), as the scaler
                           parameters may take time to estimate, but will be
                           the same across data loads/un-loads of the same
                           image.
        """
        self._image = None
        self._labels = None
        self._interpolator = None
        if unload_scaler:
            self._scaler = None

    def get_interpolator_with_current(self):
        """
        Initialize and return a ViewInterpolator object with references to this
        ImagePair's image and labels arrays. The interpolator performs linear
        and nearest interpolation of the two respectively on an arbitrary grid
        of positions in the real-scanner space (defined by the image affine).

        OBS: Does not perform interpolation in voxel space

        Returns:
            A ViewInterpolator object for the ImagePair image and labels arrays
        """
        if not self.bg_value:
            raise RuntimeError("Cannot get interpolator without a set "
                               "background value. Call self.set_bg_value "
                               "first.")
        # Standardize the bg_value, handles None and False differently from 0
        # or 0.0 - for None and False the 1st percentile of self.image is used
        if not self.predict_mode:
            labels = self.labels
        else:
            labels = None
        return ViewInterpolator(self.image, labels,
                                bg_value=self.bg_value,
                                bg_class=self.bg_class,
                                affine=self.affine)

    def set_interpolator_with_current(self):
        """
        Set an interpolator object in the ImagePair
        See get_interpolator_with_current
        """
        self._interpolator = self.get_interpolator_with_current()

    def standardize_bg_val(self, bg_value):
        """
        Standardize the bg_value, handles None and False differently from 0
        or 0.0 - for None and False the 1st percentile of self.image is used

        Args:
            bg_value: The non-standardized background value. Should be an int,
                      float, None or False

        Returns:
            A list of float(s)/int(s) image background value pr. channel
        """
        bg_value = bg_value if (bg_value is not None and bg_value is not False) \
                            else "1pct"
        return [bg_value] if not isinstance(bg_value,
                                            (list, tuple, np.ndarray)) else bg_value

    def _bg_pct_string_to_value(self, bg_pct_str):
        """
        TODO

        Args:
            bg_pct_str: TODO
        """
        bg_value = []
        for i, bg_str in enumerate(bg_pct_str):
            # assuming '<number>pct' format
            bg_pct = int(bg_str.lower().replace(" ", "").split("pct")[0])
            bg_value.append(np.percentile(self.image[..., i], bg_pct))
        self.logger.warn("Image %s: Using %s percentile BG value of %s" % (
            self.identifier, bg_pct_str, bg_value
        ),  no_print=True)
        return bg_value



import numpy as np


def expand_to_dim(values, dim):
    expanded = []
    for v in values:
        if not isinstance(v, (tuple, list, np.ndarray)):
            v = [v]
        if isinstance(v, np.ndarray):
            v = list(v)
        if len(v) != dim:
            v *= dim
        expanded.append(np.array(v))
    return expanded


def output_features(in_filter_size, padding, kernel_size, stride=1, dim=2):
    print("")
    in_filter_size, padding, kernel_size, stride = expand_to_dim([in_filter_size,
                                                                  padding,
                                                                  kernel_size,
                                                                  stride],
                                                                 dim=dim)

    return np.floor((in_filter_size + (2*padding) - kernel_size)/stride).astype(np.int) + 1


def output_feature_distance(input_feature_distance, stride, dim=2):
    input_feature_distance, stride = expand_to_dim([input_feature_distance,
                                                    stride], dim=dim)
    return input_feature_distance * stride


def output_receptive_field(input_receptive_field, kernel_size,
                           input_feature_distacne, dim=2):
    input_receptive_field, kernel_size, \
    input_feature_distacne = expand_to_dim([input_receptive_field,
                                            kernel_size,
                                            input_feature_distacne],
                                           dim=dim)
    return input_receptive_field + (kernel_size - 1) * input_feature_distacne


def output_first_feature_center(input_first_feature_center, kernel_size,
                                padding, input_feature_distance, dim=2):
    input_first_feature_center,\
    kernel_size, \
    padding, \
    input_feature_distance = expand_to_dim([input_first_feature_center,
                                            kernel_size, padding,
                                            input_feature_distance],
                                           dim=dim)

    return input_first_feature_center + ((kernel_size-1)/2 - padding) * input_feature_distance


def compute_receptive_fields(layers, verbose=False):
    input_ = layers[0]
    layers = layers[1:]
    size = input_.input.get_shape().as_list()[1:-1]
    dim = len(size)

    # Set first layer parameters
    receptive_field = 1
    jump = 1

    # Loop over all layers
    values = []
    for i, layer in enumerate(layers):
        try:
            kernel_size = layer.kernel_size
        except AttributeError as e:
            try:
                # Pooling layer?
                kernel_size = layer.pool_size
            except AttributeError:
                # Batch norm, flatten etc.
                continue
        kernel_size = np.array(kernel_size)

        # Get potential dilation rates
        try:
            dilation = np.array(layer.dilation_rate).astype(np.int)
        except AttributeError:
            dilation = np.ones(shape=[dim], dtype=np.int)
        if hasattr(layer, "dilations"):
            assert (dilation == 1).all()
            dilation = np.array(layer.dilations)
            dilation = dilation[dilation[:, 0].argmax()]

        # Get strides
        stride = layer.strides

        # Get kernel size taking into account dilation rate
        ks = kernel_size * dilation
        m = np.where(dilation > 1)
        ks[m] -= (dilation[m]-1)

        size = np.array(layer.output.get_shape().as_list()[1:-1])
        jump = output_feature_distance(jump, stride, dim)
        receptive_field = output_receptive_field(receptive_field, ks, jump, dim)

        # Add to list
        values.append((size, jump, receptive_field))

        if verbose:
            s = "\nLayer %i %s(kernel_size=%s, stride=%s, dilation=%s)" % \
                (i+1, layer.__class__.__name__, kernel_size,
                 stride, tuple(dilation))
            print(s + "\n" + "-" * (len(s)-1))
            print("Num feature:".ljust(25), size)
            print("Feature distance:".ljust(25), jump)
            print("Receptive field:".ljust(25), receptive_field)

    return values



from mpunet.evaluate import loss_functions as custom_losses
from mpunet.evaluate import metrics as custom_metrics
from mpunet.utils import ensure_list_or_tuple
from mpunet.errors.implementation_change_errors import NotSparseError
from tensorflow.keras import optimizers, losses, metrics, activations
from tensorflow_addons import optimizers as addon_optimizers
from tensorflow_addons import activations as addon_activations


# Default error message to raise with non-sparse losses or metrics passed
SPARSE_ERR = NotSparseError("This implementation now requires integer targets "
                            "as opposed to one-hot encoded targets. "
                            "All metrics and loss functions should be named "
                            "'sparse_[org_name]' to reflect this in accordance"
                            " with the naming convention of TensorFlow.keras.")


def ensure_sparse(loss_and_metric_names):
    """
    Checks that 'sparse' is a substring of each string in a list of loss and/or
    metric names. Raises NotSparseError if one or more does not contain the
    substring.
    """
    for i, m in enumerate(loss_and_metric_names):
        if "sparse" not in m:
            raise SPARSE_ERR


def _init(string_list, tf_funcs, custom_funcs, logger=None, **kwargs):
    """
    Helper for 'init_losses' or 'init_metrics'.
    Please refer to their docstrings.

    Args:
        string_list:  (list)   List of strings, each giving a name of a metric
                               or loss to use for training. The name should
                               refer to a function or class in either tf_funcs
                               or custom_funcs modules.
        tf_funcs:     (module) A Tensorflow.keras module of losses or metrics,
                               or a list of various modules to look through.
        custom_funcs: (module) A custom module or losses or metrics
        logger:       (Logger) A Logger object
        **kwargs:     (dict)   Parameters passed to all losses or metrics which
                               are represented by a class (i.e. not a function)

    Returns:
        A list of len(string_list) of initialized classes of losses or metrics
        or references to loss or metric functions.
    """
    initialized = []
    tf_funcs = ensure_list_or_tuple(tf_funcs)
    for func_or_class in ensure_list_or_tuple(string_list):
        modules_found = list(filter(None, [getattr(m, func_or_class, None)
                                           for m in tf_funcs]))
        if modules_found:
            initialized.append(modules_found[0])  # return the first found
        else:
            # Fall back to look in custom module
            initialized.append(getattr(custom_funcs, func_or_class))
    return initialized


def init_losses(loss_string_list, logger=None, **kwargs):
    """
    Takes a list of strings each naming a loss function to return. The string
    name should correspond to a function or class that is an attribute of
    either the tensorflow.keras.losses or mpunet.evaluate.losses
    modules.

    The returned values are either references to the loss functions to use, or
    initialized loss classes for some custom losses (used when the loss
    requires certain parameters to be set).

    Args:
        loss_string_list: (list)   A list of strings each naming a loss to
                                   return
        logger:           (Logger) An optional Logger object
        **kwargs:         (dict)   Parameters that will be passed to all class
                                   loss functions (i.e. not to functions)

    Returns:
        A list of length(loss_string_list) of loss functions or initialized
        classes
    """
    return _init(
        loss_string_list, losses, custom_losses, logger, **kwargs
    )


def init_metrics(metric_string_list, logger=None, **kwargs):
    """
    Same as 'init_losses', but for metrics.
    Please refer to the 'init_losses' docstring.
    """
    return _init(
        metric_string_list, metrics, custom_metrics, logger, **kwargs
    )


def init_optimizer(optimizer_string, logger=None, **kwargs):
    """
    Same as 'init_losses', but for optimizers.
    Please refer to the 'init_losses' docstring.
    """
    optimizer = _init(
        optimizer_string,
        tf_funcs=[optimizers, addon_optimizers],
        custom_funcs=None,
        logger=logger
    )[0]
    return optimizer(**kwargs)


def init_activation(activation_string, logger=None, **kwargs):
    """
    Same as 'init_losses', but for optimizers.
    Please refer to the 'init_losses' docstring.
    """
    activation = _init(
        activation_string,
        tf_funcs=[activations, addon_activations],
        custom_funcs=None,
        logger=logger
    )[0]
    return activation



import os
from glob import glob
from argparse import ArgumentParser


def get_argparser():
    parser = ArgumentParser(description='Replace .nii(.gz) files matching a '
                                        'GLOB regex in-place to keep only 1 '
                                        'out of multiple channels')
    parser.add_argument("-d", type=str, default="./",
                        help='root dir')
    parser.add_argument("-e", type=str, required=True,
                        help="GLOB regex matching files")
    parser.add_argument("-c", type=int, default=0, help="Keep only this channel")

    return parser


def entry_func(args=None):
    args = vars(get_argparser().parse_args(args))
    base_path = os.path.abspath(args["d"])
    match = os.path.join(base_path, args["e"])
    c = args["c"]

    files = glob(match, recursive=True)

    if files:
        print("Matching %i files:" % len(files))
        for f in files:
            print(f)

        answer = input("\n[OBS] Replace IN-PLACE with files "
                       "containing only channel %i? (y/N) " % c)

        if answer.lower() in ("y", "yes"):
            import nibabel as nib
            import numpy as np

            for f in files:
                im = nib.load(f)
                new_im = im.get_data()[..., c].astype(np.float32)
                nib.save(nib.Nifti1Image(new_im, affine=im.affine), f)
    else:
        print("No matches to GLOB: '%s'" % match)


if __name__ == "__main__":
    entry_func()



import os
import inspect
from mpunet.utils.decorators import accepts
from mpunet.utils import create_folders
from contextlib import contextmanager
from threading import Lock


class Logger(object):
    def __init__(self, base_path, print_to_screen=True, active_file="log.txt",
                 overwrite_existing=False, append_existing=False,
                 print_calling_method=True, no_sub_folder=False,
                 log_prefix="", warnings_file="warnings"):
        self.base_path = os.path.abspath(base_path)
        if not no_sub_folder:
            self.path = os.path.join(self.base_path, "logs")
        else:
            self.path = self.base_path
        create_folders([self.path])
        if overwrite_existing and append_existing:
            raise ValueError("Cannot set both 'overwrite_existing' and "
                             "'append_existing' to True.")
        self.overwrite_existing = overwrite_existing
        self.append_existing = append_existing
        self._enabled = True

        # Get built in print function
        # (if overwritten globally, Logger still maintains a reference to the
        # true print function)
        self.print_f = __builtins__["print"]

        # Print options
        self.separator = "-" * 80
        self.print_to_screen = print_to_screen
        self.print_calling_method = print_calling_method

        # Set paths to log files
        self.log_files = {}
        self.currently_logging = {}
        self.prefix = "" if log_prefix is None else str(log_prefix)
        self.active_log_file = active_file
        self.warnings_file = warnings_file

        # For using the logger from multiple threads
        self.lock = Lock()

    def __repr__(self):
        return "<mpunet.logging.Logger object>"

    def __str__(self):
        return "Logger(base_path=%s, print_to_screen=%s, " \
               "overwrite_existing=%s, append_existing=%s)" \
               % (self.base_path, self.print_to_screen,
                  self.overwrite_existing, self.append_existing)

    def new_log_file(self, filename):
        file_path = os.path.join(self.path, filename)

        if os.path.exists(file_path):
            if self.overwrite_existing:
                os.remove(file_path)
            elif not self.append_existing:
                raise OSError("Logging path: %s already exists. "
                              "Initialize Logger with overwrite_existing=True "
                              "or append_existing=True to overwrite or continue"
                              " writing to the existing file." % file_path)

        self.log_files[filename] = file_path
        self.currently_logging[filename] = None

        # Add reference to model folder in log
        ref = "Log for model in: %s" % self.base_path
        self._add_to_log(ref, no_print=True)

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = bool(value)

    @contextmanager
    def disabled_in_context(self):
        self.enabled = False
        yield self
        self.enabled = True

    @property
    def print_to_screen(self):
        return self._print_to_screen

    @print_to_screen.setter
    @accepts(bool)
    def print_to_screen(self, value):
        self._print_to_screen = value

    @property
    def print_calling_method(self):
        return self._print_calling_method

    @print_calling_method.setter
    @accepts(bool)
    def print_calling_method(self, value):
        self._print_calling_method = value

    @property
    def log(self):
        with open(self.log_files[self.active_log_file], "r") as log_f:
            return log_f.read()

    @property
    def active_log_file(self):
        return self._active_log_file

    @active_log_file.setter
    @accepts(str)
    def active_log_file(self, file_name):
        if self.prefix:
            file_name = file_name.replace(self.prefix + "_", "")
            file_name = self.prefix.rstrip("_") + "_" + file_name
        file_name, ext = os.path.splitext(file_name)
        file_name = "%s%s" % (file_name, ext or ".txt")
        self._active_log_file = file_name
        if file_name not in self.log_files:
            self.new_log_file(file_name)

    def _add_to_log(self, *args, no_print=False, **kwargs):
        if self.print_to_screen and not no_print:
            self.print_f(*args, **kwargs)

        with open(self.log_files[self.active_log_file], "a") as log_file:
            self.print_f(*args, file=log_file, **kwargs)

    def _log(self, caller, print_calling_owerwrite=None, *args, **kwargs):
        if caller != self.currently_logging[self.active_log_file]:
            self.currently_logging[self.active_log_file] = caller
            if print_calling_owerwrite is not None:
                print_calling = print_calling_owerwrite
            else:
                print_calling = self.print_calling_method
            if print_calling:
                self._add_to_log("%s\n>>> Logged by: %s" % (self.separator,
                                                            self.currently_logging[self.active_log_file]))
        self._add_to_log(*args, **kwargs)

    def __call__(self, *args,
                 print_calling_method=None,
                 out_file=None,
                 **kwargs):
        if not self.enabled:
            return None
        with self.lock:
            cur_file = self.active_log_file
            self.active_log_file = out_file or cur_file
            caller = inspect.stack()[1]
            caller = "'%s' in '%s'" % (caller[3], caller[1].rpartition("/")[2])
            self._log(caller, print_calling_method, *args, **kwargs)
            self.active_log_file = cur_file

    def warn(self, *args, **kwargs):
        self.__call__("[WARNING]", *args,
                      print_calling_method=False,
                      out_file=self.warnings_file,
                      **kwargs)



"""
TODO: Add description
"""

import os
import sys
import numpy as np
import pandas as pd
from glob import glob
from argparse import ArgumentParser


def get_argparser():
    parser = ArgumentParser(description='Fit a mpunet model defined in a project folder. '
                                        'Invoke "init_project" to start a new project.')
    parser.add_argument("--root_dir", type=str, default="./",
                        help='Starting point from which csv '
                             'folders will be searched for.')
    parser.add_argument("--pred_subdir", type=str, default="predictions",
                        help="Subdirectory storing the 'csv' subfolder.")
    parser.add_argument("--round", default=3, type=int)
    return parser


def print_mj_or_detailed(df, round_):
    classes = df.get("class")
    if classes is not None:
        ddf = df.drop(["class"], axis=1)
    else:
        classes = np.arange(1, df.shape[0]+1)
        ddf = df
    means = np.nanmean(ddf.values, axis=1)
    stds = np.nanstd(ddf.values, axis=1)
    mins = np.nanmin(ddf.values, axis=1)
    maxs = np.nanmax(ddf.values, axis=1)
    N = np.sum(~np.isnan(ddf.values), axis=1)

    print("\nPer class:\n--------------------------------")
    print_df = pd.DataFrame(
        {c: [m, std, min, max, n] for c, m, std, min, max, n in zip(classes, means, stds, mins, maxs, N)}).T
    print_df.columns = ["Mean dice by class", "+/- STD", "min", "max", "N"]

    print(np.round(print_df, round_))
    print("\nOverall mean: %s +- %s" % (
        np.nanmean(ddf.values).round(round_),
        np.nanstd(ddf.values).round(round_)))
    print("--------------------------------")


def print_res(df, round_):
    to_drop = ("Unnamed: 0", "identifier", "MJ")
    for d in to_drop:
        try:
            df = df.drop([d], axis=1)
        except KeyError:
            pass
    print("\nBy views:\n--------------------------------")

    longest = max([len(v) for v in df])
    for v in df:
        mean = np.nanmean(df[v]).round(round_)
        print(("%s" % v).ljust(longest + 7) + "%s" % mean)
    print("--------------------------------")


def print_results(results, folder, round_):
    print("\n[***] SUMMARY REPORT FOR FOLDER [***]\n%s\n" % folder)

    for file_ in results:
        df = results[file_]

        if file_ in ("MJ.csv", "detailed.csv"):
            print_mj_or_detailed(df, round_)
        elif file_ == "results.csv":
            print_res(df, round_)
        else:
            raise ValueError("Unknown file type '%s'" % file_)


def parse_folder(folder, look_for=("MJ", "results", "detailed")):
    import pandas as pd
    dfs = {}
    files = os.listdir(folder)
    for tag in look_for:
        for f in files:
            if tag in f:
                dfs[f] = pd.read_csv(os.path.join(folder, f))

    return dfs


def pool_results(results):
    import pandas as pd
    pooled = {}
    for folder in results:
        for file_ in results[folder]:
            df = results[folder][file_]
            if pooled.get(file_) is not None:
                if file_ in ("MJ.csv", "detailed.csv"):
                    pooled[file_] = pd.merge(pooled[file_], df,
                                             left_index=True,
                                             right_index=True)
                elif file_ == "results.csv":
                    pooled[file_] = pd.concat([pooled[file_], df],
                                              axis=0,
                                              sort=True)
                else:
                    raise ValueError("Unknown file type '%s'" % file_)
            else:
                pooled[file_] = df
    return pooled


def parse_results(csv_folders, round_=3):
    results = {}
    for folder in csv_folders:
        results[folder] = parse_folder(folder)
    results = pool_results(results)
    if len(csv_folders) > 1:
        folder = "Pool of %i folders" % len(csv_folders)
    else:
        folder = csv_folders[0]

    print_results(results, folder, round_)


def entry_func(args=None):

    args = vars(get_argparser().parse_args(args))
    dir_ = os.path.abspath(args["root_dir"])
    p_dir = args["pred_subdir"]
    round_ = args["round"]

    # Get folder/folders - 3 levels possible
    csv_folders = glob("%s/csv" % dir_)
    if not csv_folders:
        csv_folders = glob("%s/%s/csv/" % (dir_, p_dir))
        if not csv_folders:
            csv_folders = glob(dir_ + "/**/%s/csv/" % p_dir)
    if not csv_folders:
        print("Could not locate result csv files.")
        sys.exit(0)
    else:
        print("Found %i 'csv' folders under prediction "
              "dirs '%s':" % (len(csv_folders), p_dir))
        csv_folders.sort()
        for d in csv_folders:
            print("-- " + os.path.abspath(d))
        in_ = input("\nCorrect? (Y/n) ")
        if in_.lower() not in ("n", "no"):
            parse_results(csv_folders, round_)


if __name__ == "__main__":
    entry_func()



import numpy as np

from mpunet.utils.fusion import map_real_space_pred
from mpunet.evaluate import dice_all


def stack_collections(points_collection, targets_collection):
    """

    Args:
        points_collection:
        targets_collection:

    Returns:

    """
    if len(points_collection) == 1 and len(targets_collection) == 1:
        return points_collection[0], targets_collection[0]
    n_points = sum([x.shape[0] for x in points_collection])
    n_views, n_classes = points_collection[0].shape[1:]

    X = np.empty(shape=(n_points, n_views, n_classes),
                 dtype=points_collection[0].dtype)
    y = np.empty(shape=(n_points, 1),
                 dtype=targets_collection[0].dtype)

    c = 0
    len_collection = len(points_collection)
    for i in range(len_collection):
        print("  %i/%i" % (i + 1, len_collection),
              end="\r", flush=True)
        Xs = points_collection.pop()
        X[c:c + len(Xs)] = Xs
        y[c:c + len(Xs)] = targets_collection.pop()
        c += len(Xs)
    print("")
    return X, y


def predict_and_map(model, seq, image, view, batch_size=None,
                    voxel_grid_real_space=None, targets=None, eval_prob=1.0,
                    n_planes='same+20'):
    """


    Args:
        model:
        seq:
        image:
        view:
        batch_size:
        voxel_grid_real_space:
        targets:
        n_planes:

    Returns:

    """

    # Sample planes from the image at grid_real_space grid
    # in real space (scanner RAS) coordinates.
    X, y, grid, inv_basis = seq.get_view_from(image, view, n_planes=n_planes)

    # Predict on volume using model
    bs = seq.batch_size if batch_size is None else batch_size
    from mpunet.utils.fusion import predict_volume
    pred = predict_volume(model, X, axis=2, batch_size=bs)

    # Map the real space coordiante predictions to nearest
    # real space coordinates defined on voxel grid
    if voxel_grid_real_space is None:
        from mpunet.interpolation.sample_grid import get_voxel_grid_real_space
        voxel_grid_real_space = get_voxel_grid_real_space(image)

    # Map the predicted volume to real space
    mapped = map_real_space_pred(pred, grid, inv_basis, voxel_grid_real_space)

    # Print dice scores
    if targets is not None and np.random.rand(1)[0] <= eval_prob:
        print("Computing evaluations...")
        print("View dice scores:   ", dice_all(y, pred.argmax(-1),
                                               ignore_zero=False))
        print("Mapped dice scores: ", dice_all(targets,
                                               mapped.argmax(-1).reshape(-1, 1),
                                               ignore_zero=False))
    else:
        print("-- Skipping evaluation")

    return mapped



import numpy as np
from mpunet.image.queue import BaseQueue, LoadingPool
from queue import Queue, Empty
from contextlib import contextmanager


class LimitationQueue(BaseQueue):
    """
    Implements an image loading queue.
    Stores a reference to a ImagePairLoader object.

    Using the methods get_random_image method, this method tracks the number of
    times a ImagePair object has been accessed, and when exceeding a threshold,
    unloads it and loads a random ImagePair from the dataset.
    """
    def __init__(self,
                 dataset,
                 max_loaded=25,
                 num_access_before_reload=50,
                 preload_now=True,
                 await_preload=True,
                 loading_pool=None,
                 n_load_jobs=5,
                 logger=None,
                 **kwargs):
        """
        Initialize a LoadQueue object from a ImagePairLoader object

        Args:
            dataset:                    (list) A ImagePairLoader object
            max_loaded:                 (int)  Number of ImagePair objects in
                                               a dataset that will be loaded
                                               at a given time.
            num_access_before_reload:   (int)  Number of times a ImagePair obj
                                               can be accessed be
                                               get_random_image or
                                               a unload is invoked and a new
                                               data point is loaded.
            preload_now:   TODO
            image_loader:  TODO
            n_load_jobs:   TODO
            logger:        TODO
        """
        super(LimitationQueue, self).__init__(
            dataset=dataset,
            logger=logger
        )
        self.max_loaded = min(max_loaded or len(dataset), len(dataset))
        self.num_access_before_reload = num_access_before_reload or 50

        # Queues of loaded and non-loaded objects
        self.loaded_queue = Queue(maxsize=self.max_loaded)
        self.non_loaded_queue = Queue(maxsize=len(dataset))

        # Fill non-loaded queue in random order
        inds = np.arange(len(dataset))
        np.random.shuffle(inds)
        for i in inds:
            self.non_loaded_queue.put(self.dataset.images[i])

        # Setup load thread pool
        self.loading_pool = loading_pool or LoadingPool(
            n_threads=n_load_jobs
        )
        # Register this dataset to become updated with new loaded images
        # from the StudyLoader thread.
        self.loading_pool.register_dataset(
            dataset_id=self.dataset.identifier,
            load_put_function=self._add_loaded_to_queue,
            error_put_function=self._load_error_callback,
        )

        # Increment counters to random off-set points for the first images
        self.max_offset = int(self.num_access_before_reload * 0.75)
        self.n_offset = self.max_loaded

        self.logger("'Limitation' queue created:\n"
                    "  Dataset:      {}\n"
                    "  Max loaded:   {}\n"
                    "  Reload limit: {}".format(
            self.dataset, self.max_loaded, self.num_access_before_reload
        ))
        if preload_now:
            # Load specified number of obj and populate access count dict
            self.preload(await_preload)

    def preload(self, await_preload=True):
        """
        TODO

        Returns:

        """
        # Set the number of loaded objects to 'max_loaded_per_dataset'
        self.logger("Adding {} ImagePair objects from "
                    "{} to load queue".format(self.max_loaded,
                                              self.dataset.identifier))
        if self.dataset.n_loaded != 0 or self.loaded_queue.qsize() != 0:
            raise RuntimeError("Dataset {} seems to have already been "
                               "loaded. Do not load any data before "
                               "passing the ImagePairLoader object "
                               "to the queue class. Only call "
                               "LoadQueue.preload once."
                               "".format(self.dataset.identifier))
        self._add_images_to_load_queue(num=self.max_loaded)
        if await_preload:
            self.logger("... awaiting preload")
            self.loading_pool.join()
            self.logger("Preload complete.")

    def load_queue_too_full(self, max_fraction=0.33):
        return self.loading_pool.qsize() > \
               self.loading_pool.maxsize*max_fraction

    def _warn_access_limit(self, min_fraction=0.10):
        qsize = self.loaded_queue.qsize()
        if qsize == 0:
            self.logger.warn("Study ID queue for dataset {} seems to"
                             " block. This might indicate a data loading "
                             "bottleneck.".format(self.dataset.identifier))
        elif qsize <= self.max_loaded*min_fraction:
            self.logger.warn("Dataset {}: Loaded queue in too empty "
                             "(qsize={}, maxsize={})"
                             .format(self.dataset.identifier, qsize,
                                     self.max_loaded))

    @contextmanager
    def get_image_by_id(self, image_id):
        raise NotImplementedError

    @contextmanager
    def get_image_by_idx(self, image_idx):
        raise NotImplementedError

    @contextmanager
    def get_random_image(self):
        """
        TODO

        Returns:

        """
        with self.loading_pool.thread_lock:
            self._warn_access_limit()
        timeout_s = 5
        try:
            image_pair, n_accesses = self.loaded_queue.get(timeout=timeout_s)
        except Empty as e:
            raise Empty("Could not get ImagePair from dataset {} with "
                        "timeout of {} seconds. Consider increasing the "
                        "number of load threads / max loaded per dataset /"
                        " access threshold".format(self.dataset.identifier,
                                                   timeout_s)) from e
        try:
            yield image_pair
        finally:
            self._release_image(image_pair, n_accesses)

    def _add_images_to_load_queue(self, num=1):
        """
        TODO

        Args:
            num:

        Returns:

        """
        for _ in range(num):
            image = self.non_loaded_queue.get_nowait()  # Should never block!
            if image.is_loaded:
                raise RuntimeWarning("Image {} in dataset {} seems to be "
                                     "already loaded, but it was fetched from "
                                     "the self.non_loaded_queue queue. This "
                                     "could be an implementation error!"
                                     "".format(image.identifier,
                                               self.dataset.identifier))
            self.loading_pool.add_image_to_load_queue(image, self.dataset.identifier)

    def _add_loaded_to_queue(self, image_pair):
        """

        Args:
            image_id:

        Returns:

        """
        if self.n_offset >= 0:
            offset = np.random.randint(0, self.max_offset)
            self.n_offset -= 1
        else:
            offset = 0
        self.loaded_queue.put((image_pair, offset))

    def _load_error_callback(self, image_pair, *args, **kwargs):
        self.logger.warn("Load error on image {}".format(image_pair))
        self._add_images_to_load_queue(num=1)

    def _release_image(self, image_pair, n_accesses):
        """
        TODO

        Args:
            image_pair:
            n_accesses:

        Returns:

        """
        if n_accesses >= self.num_access_before_reload:
            # Unload, add to unloaded queue, start loading new image
            image_pair.unload()
            self.non_loaded_queue.put(image_pair)
            self._add_images_to_load_queue(num=1)
        else:
            self.loaded_queue.put((image_pair, n_accesses+1))



"""
Mathias Perslev
MSc Bioinformatics

University of Copenhagen
November 2017
"""

import numpy as np
import tensorflow as tf


def dice(y_true, y_pred, smooth=1.0):
    """
    Calculates the Soerensen dice coefficient between two binary sets
    """
    # Flatten and bool the ground truth and predicted labels
    s1 = np.array(y_true).flatten().astype(np.bool)
    s2 = np.array(y_pred).flatten().astype(np.bool)

    # Calculate dice
    return (smooth + 2 * np.logical_and(s1, s2).sum()) \
           / (smooth + s1.sum() + s2.sum())


def dice_all(y_true, y_pred, smooth=1.0, n_classes=None, ignore_zero=True,
             skip_if_no_y=False):
    """
    Calculates the Soerensen dice coefficients for all unique classes
    """
    # Get array of unique classes in true label array
    if n_classes is None:
        classes = np.unique(y_true)
    else:
        classes = np.arange(max(2, n_classes))
    # Ignore background class?
    if ignore_zero:
        classes = classes[np.where(classes != 0)]

    # Calculate dice for all targets
    dice_coeffs = np.empty(shape=classes.shape, dtype=np.float32)
    dice_coeffs.fill(np.nan)
    for idx, _class in enumerate(classes):
        s1 = y_true == _class
        if skip_if_no_y and not np.any(s1):
            continue
        s2 = y_pred == _class

        if np.any(s1) or np.any(s2):
            d = dice(s1, s2, smooth=smooth)
            dice_coeffs[idx] = d
    return dice_coeffs


def class_wise_kappa(true, pred, n_classes=None, ignore_zero=True):
    from sklearn.metrics import cohen_kappa_score
    if n_classes is None:
        classes = np.unique(true)
    else:
        classes = np.arange(max(2, n_classes))
    # Ignore background class?
    if ignore_zero:
        classes = classes[np.where(classes != 0)]

    # Calculate kappa for all targets
    kappa_scores = np.empty(shape=classes.shape, dtype=np.float32)
    kappa_scores.fill(np.nan)
    for idx, _class in enumerate(classes):
        s1 = true == _class
        s2 = pred == _class

        if np.any(s1) or np.any(s2):
            kappa_scores[idx] = cohen_kappa_score(s1, s2)
    return kappa_scores


def one_class_dice(y_true, y_pred, smooth=1.0):
    # Predict
    y_pred = tf.cast(y_pred > 0.5, tf.float32)

    return (smooth + 2.0 * tf.reduce_sum(y_true * y_pred)) / (smooth + tf.reduce_sum(y_true) + tf.reduce_sum(y_pred))


def sparse_fg_recall(y_true, y_pred, bg_class=0):
    # Get MAP estimates
    y_true = tf.cast(tf.reshape(y_true, [-1]), tf.int32)
    y_pred = tf.cast(tf.reshape(tf.argmax(y_pred, axis=-1), [-1]), tf.int32)

    # Remove background
    mask = tf.not_equal(y_true, bg_class)
    y_true = tf.boolean_mask(y_true, mask)
    y_pred = tf.boolean_mask(y_pred, mask)

    return tf.reduce_mean(tf.cast(tf.equal(y_true, y_pred), tf.float32))


def sparse_mean_fg_f1(y_true, y_pred):
    y_pred = tf.argmax(y_pred, axis=-1)

    # Get confusion matrix
    cm = tf.confusion_matrix(tf.reshape(y_true, [-1]),
                             tf.reshape(y_pred, [-1]))

    # Get precisions
    TP = tf.diag_part(cm)
    precisions = TP / tf.reduce_sum(cm, axis=0)

    # Get recalls
    TP = tf.diag_part(cm)
    recalls = TP / tf.reduce_sum(cm, axis=1)

    # Get F1s
    f1s = (2 * precisions * recalls) / (precisions + recalls)

    return tf.reduce_mean(f1s[1:])


def sparse_mean_fg_precision(y_true, y_pred):
    y_pred = tf.argmax(y_pred, axis=-1)

    # Get confusion matrix
    cm = tf.confusion_matrix(tf.reshape(y_true, [-1]),
                             tf.reshape(y_pred, [-1]))

    # Get precisions
    TP = tf.diag_part(cm)
    precisions = TP / tf.reduce_sum(cm, axis=0)

    return tf.reduce_mean(precisions[1:])


def sparse_mean_fg_recall(y_true, y_pred):
    y_pred = tf.argmax(y_pred, axis=-1)

    # Get confusion matrix
    cm = tf.confusion_matrix(tf.reshape(y_true, [-1]),
                             tf.reshape(y_pred, [-1]))

    # Get precisions
    TP = tf.diag_part(cm)
    recalls = TP / tf.reduce_sum(cm, axis=1)

    return tf.reduce_mean(recalls[1:])


def sparse_fg_precision(y_true, y_pred, bg_class=0):
    # Get MAP estimates
    y_true = tf.cast(tf.reshape(y_true, [-1]), tf.int32)
    y_pred = tf.cast(tf.reshape(tf.argmax(y_pred, axis=-1), [-1]), tf.int32)

    # Remove background
    mask = tf.not_equal(y_pred, bg_class)
    y_true = tf.boolean_mask(y_true, mask)
    y_pred = tf.boolean_mask(y_pred, mask)

    return tf.reduce_mean(tf.cast(tf.equal(y_true, y_pred), tf.float32))



import unittest
import os
import shutil


class TestSetup(unittest.TestCase):

    def test_mp_on_path(self):
        """ Tests if the mp script is reachable by command 'mp' """
        path = shutil.which("mp")
        self.assertIsNot(path, None)
        self.assertTrue(os.path.exists(path))



import matplotlib
matplotlib.use('agg')
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
from mpunet.logging import ScreenLogger


def save_images(train, val, out_dir, logger):
    logger = logger or ScreenLogger()
    # Write a few images to disk
    im_path = out_dir
    if not os.path.exists(im_path):
        os.mkdir(im_path)

    # Sample two batches of training and (if passed) validation data
    training = [train[0], train[1]]
    if val is not None and len(val) != 0:
        validation = [val[0], val[1]]
        val_bs = val.batch_size
    else:
        validation = [None] * len(training)
        val_bs = 0

    logger("Saving %i sample images in '<project_dir>/images' folder"
           % ((train.batch_size + val_bs) * 2))
    for rr, (train_batch, val_batch) in enumerate(zip(training, validation)):
        for k, temp in enumerate((train_batch, val_batch)):
            if temp is None:
                # No validation data
                continue
            X, Y, W = temp
            for i, (xx, yy, ww) in enumerate(zip(X, Y, W)):
                yy = yy.reshape(xx.shape[:-1] + (yy.shape[-1],))

                # Make figure
                fig = plt.figure(figsize=(10, 4))
                ax1 = fig.add_subplot(121)
                ax2 = fig.add_subplot(122)

                # Plot image and overlayed labels
                chnl, view, _ = imshow_with_label_overlay(ax1, xx, yy)

                # Plot histogram
                ax2.hist(xx.flatten(), bins=200)

                # Set labels
                ax1.set_title("Channel %i - Axis %i - "
                              "Weight %.3f" % (chnl, view, ww), size=18)

                # Get path
                out_path = im_path + "/%s%i.png" % ("train" if k == 0 else
                                                    "val", len(X) * rr + i)

                with np.testing.suppress_warnings() as sup:
                    sup.filter(UserWarning)
                    fig.savefig(out_path)
                plt.close(fig)


def imshow(ax, image, channel=None, axis=None, slice=None, cmap="gray"):
    """
    Imshow an image of dim 2 or dim 3

    Args:
        ax:
        image:
        channel:
        axis:
        slice:
        cmap:

    Returns:

    """
    # Get channel to plot
    if channel is None:
        channel = np.random.randint(0, image.shape[-1], 1)[0]

    # Get image channel
    image = image[..., channel]
    img_dims = image.ndim

    if img_dims == 3 and axis is None:
        shape = np.array(image.shape[:-1])
        if np.all(shape == shape[0]):
            axis = np.random.randint(0, shape.ndim, 1)[0]
        else:
            axis = np.argmin(shape)
    elif axis is None:
        axis = 0

    # Move the chosen axis forward
    image = np.moveaxis(image, axis, 0)

    if img_dims == 3:
        if slice is None:
            # Get a random slice around the middle of the axis
            slice = np.random.randint(0 + (len(image) // 3),
                                      len(image) - (len(image) // 3), 1)[0]
        im_slice = image[slice]
    else:
        im_slice = image

    # Imshow
    ax.imshow(im_slice, cmap=cmap)

    return channel, axis, slice


def imshow_with_label_overlay(ax, image, label, channel=None, axis=None,
                              slice=None, im_cmap="gray", lab_cmap=None,
                              lab_alpha=0.7):
    """
    Imshow an image of dim 2 or dim 3 with labels overlayed on a single ax

    Args:
        ax:
        image:
        label:
        channel:
        axis:
        im_cmap:
        lab_cmap:
        lab_alpha:

    Returns:

    """
    from mpunet.utils import pred_to_class

    # Plot the image
    channel, axis, slice = imshow(ax, image, channel, axis, slice, im_cmap)

    # Get int labels if needed, otherwise returns identity
    img_dims = image.ndim-1
    label = pred_to_class(label, img_dims=img_dims, has_batch_dim=False)

    # Move the chosen axis forward
    label = np.moveaxis(label, axis, 0)

    if img_dims == 3:
        lab_slice = label[slice]
    else:
        lab_slice = label

    # Overlay masked label image
    masked_lab = np.ma.masked_where(lab_slice == 0, lab_slice)

    # Imshow labels
    ax.imshow(masked_lab, alpha=lab_alpha, cmap=lab_cmap)

    return channel, axis, slice


def plot_all_training_curves(glob_path, out_path, raise_error=False,
                             logger=None, **kwargs):
    logger = logger or ScreenLogger()
    try:
        from glob import glob
        paths = glob(glob_path)
        if not paths:
            raise OSError("File pattern {} gave none or too many matches " \
                          "({})".format(glob_path, paths))
        out_folder = os.path.split(out_path)[0]
        for p in paths:
            if len(paths) > 1:
                # Set unique names
                uniq = os.path.splitext(os.path.split(p)[-1])[0]
                f_name = uniq + "_" + os.path.split(out_path)[-1]
                save_path = os.path.join(out_folder, f_name)
            else:
                save_path = out_path
            plot_training_curves(p, save_path, **kwargs)
    except Exception as e:
        s = "Could not plot training curves. ({})".format(e)
        if raise_error:
            raise RuntimeError(s) from e
        else:
            logger.warn(s)


def plot_training_curves(csv_path, save_path, logy=False,
                         exclude=("learning_rate", "epoch", "loss",
                                  "epoch_minutes", "train_hours",
                                  'memory_usage_gib'),
                         include_regex=None):
    # Read CSV file
    df = pd.read_csv(csv_path)

    # Prepare plot
    fig = plt.figure(figsize=(12, 12))
    ax1 = fig.add_subplot(311)

    # Get epoch, training and validation loss vectors
    epochs = df["epoch"] + 1
    train_loss = df["loss"]
    val_loss = df.get("val_loss")

    if logy:
        train_loss = np.log10(train_loss)
        if val_loss is not None:
            val_loss = np.log10(val_loss)

    # Plot
    ax1.plot(epochs, train_loss, lw=3, color="darkblue", label="Training loss")
    if val_loss is not None:
        ax1.plot(epochs, val_loss, lw=3, color="darkred", label="Validation loss")

    # Add legend, labels and title
    leg = ax1.legend(loc=0)
    leg.get_frame().set_linewidth(0)
    ax1.set_xlabel("Epoch", size=16)
    ax1.set_ylabel("Loss" if not logy else "$\log_{10}$(Loss)", size=16)
    ax1.set_title("Training %sloss" % ("and validation " if val_loss is not None else ""), size=20)

    # Make second plot
    ax2 = fig.add_subplot(312)

    # Get all other columns, optionally only if matching 'include_regex'
    import re
    include_regex = re.compile(include_regex or ".*")

    plotted = 0
    for col in df.columns:
        if any([s in col for s in exclude[1:]]) or col == "lr":
            continue
        elif not re.match(include_regex, col):
            continue
        else:
            plotted += 1
            ax2.plot(epochs, df[col], label=col, lw=2)

    # Add legend, labels and title
    if plotted <= 8:
        # Otherwise it takes up all the space
        leg = ax2.legend(loc=0, ncol=int(np.ceil(plotted/5)))
        leg.get_frame().set_linewidth(0)
    ax2.set_xlabel("Epoch", size=16)
    ax2.set_ylabel("Metric", size=16)
    ax2.set_title("Training and validation metrics", size=20)

    # Plot learning rate
    lr = df.get("lr")
    if lr is None:
        lr = df.get("learning_rate")
    if lr is not None:
        ax3 = fig.add_subplot(313)
        ax3.step(epochs, lr)
        ax3.set_xlabel("Epoch", size=16)
        ax3.set_ylabel("Learning Rate", size=16)
        ax3.set_title("Learning Rate", size=20)

    fig.tight_layout()
    fig.savefig(save_path)
    plt.close(fig.number)


def plot_views(views, out_path):
    from mpl_toolkits.mplot3d import Axes3D

    # Create figure, 3D
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111, projection="3d")

    # Set axes
    ax.set_xlim(-0.6, 0.6)
    ax.set_ylim(-0.6, 0.6)
    ax.set_zlim(-0.6, 0.6)

    # Plot unit sphere
    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)

    x = np.outer(np.cos(u), np.sin(v))
    y = np.outer(np.sin(u), np.sin(v))
    z = np.outer(np.ones(np.size(u)), np.cos(v))
    z_f = np.zeros_like(y)

    ax.plot_surface(x, y, z, alpha=0.1, color="darkgray")
    ax.plot_surface(x, y, z_f, alpha=0.1, color="black")

    # Plot basis axes
    ax.plot([-1, 1], [0, 0], [0, 0], color="blue", linewidth=0.7)
    ax.plot([0, 0], [-1, 1], [0, 0], color="red", linewidth=0.7)
    ax.plot([0, 0], [0, 0], [-1, 1], color="green", linewidth=0.7)

    # Plot views
    for v in views:
        c = np.random.rand(3,)
        ax.scatter(*v, s=50, color=c)
        ax.scatter(*v, s=50, color=c)
        ax.plot([0, v[0]], [0, v[1]], [0, v[2]], color=c, linewidth=2)

        # Plot dashed line to XY plane
        ax.plot([v[0], v[0]], [v[1], v[1]], [0, v[2]], color="gray",
                linewidth=1, linestyle="--")

    ax.view_init(30, -45)
    ax.grid(False)
    ax.axis("off")
    fig.savefig(out_path)



from .scaling import apply_scaling, get_scaler
from .input_prep import reshape_add_axis, one_hot_encode_y


def get_preprocessing_func(model):
    """
    Takes a model name (string) and returns a preparation function.

    Args:
        model: String representation of a mpunet.models model class

    Returns:
        A mpunet.preprocessing.data_preparation_funcs function
    """
    from mpunet.models import PREPARATION_FUNCS
    if model in PREPARATION_FUNCS:
        return PREPARATION_FUNCS[model]
    else:
        raise ValueError("Unsupported model type '%s'. "
                         "Supported models: %s" % (model,
                                                   PREPARATION_FUNCS.keys()))



from mpunet.sequences import BaseSequence
from mpunet.logging import ScreenLogger
from mpunet.interpolation.linalg import mgrid_to_points
import numpy as np


def center_expand(im, target_dim, bg_value, random=True):
    if im.ndim == 4:
        cim = np.empty(shape=(target_dim, target_dim, target_dim, im.shape[-1]),
                       dtype=im.dtype)
    else:
        cim = np.empty(shape=(target_dim, target_dim, target_dim),
                       dtype=im.dtype)
    cim.fill(bg_value)

    # Calculate size difference and starting positions in new volume
    diff = np.asarray(cim.shape[:3]) - im.shape[:3]
    if random and np.any(diff):
        start = [np.random.randint(0, max(d, 1), 1)[0] for d in diff]
    else:
        start = diff//2

    cim[start[0]:start[0] + im.shape[0],
        start[1]:start[1] + im.shape[1],
        start[2]:start[2] + im.shape[2]] = im

    return cim


class PatchSequence3D(BaseSequence):
    def __init__(self, image_pair_loader, dim, n_classes, batch_size, is_validation=False,
                 label_crop=None, fg_batch_fraction=0.33, logger=None, bg_val=0.,
                 no_log=False, **kwargs):
        super().__init__()

        # Set logger or default print
        self.logger = logger or ScreenLogger()
        self.image_pair_loader = image_pair_loader

        # Box dimension and image dim
        self.dim = dim

        # Various attributes
        self.n_classes = n_classes
        self.label_crop = label_crop
        self.is_validation = is_validation
        self.batch_size = batch_size
        self.bg_value = bg_val

        # How many foreground slices should be in each batch?
        self.fg_batch_fraction = fg_batch_fraction

        # Foreground label settings
        self.fg_classes = np.arange(1, self.n_classes)
        if self.fg_classes.shape[0] == 0:
            self.fg_classes = 1

        if not is_validation and not no_log:
            self.log()

    def __len__(self):
        return int(np.ceil(self.n_samples / self.batch_size))

    @property
    def n_samples(self):
        # In order to not load all the images, we manually specify the epoch
        # length in trainer.py
        return np.inf

    @property
    def batch_size(self):
        return self._batch_size

    @batch_size.setter
    def batch_size(self, value):
        if value < 0:
            raise ValueError("Invalid batch size of %i" % value)
        self._batch_size = value

    @property
    def n_fg_slices(self):
        if self.is_validation:
            return self.batch_size
        else:
            return int(np.ceil(self.batch_size * self.fg_batch_fraction))

    def get_N_random_patches_from(self, image, N):
        if N > 0:
            # Sample N patches from X
            X = image.image
            for i in range(N):
                xc, yc, zc = self.get_random_box_coords(X)
                patch = X[xc:xc + self.dim, yc:yc + self.dim, zc:zc + self.dim]
                yield image.scaler.transform(patch), (xc, yc, zc)
        else:
            return []

    def get_base_patches(self, image):
        X = image.image

        # Calculate positions
        sample_space = np.asarray([max(i, self.dim) for i in image.shape[:3]])
        d = (sample_space - self.dim)
        min_cov = [np.ceil(sample_space[i]/self.dim).astype(np.int) for i in range(3)]
        ds = [np.linspace(0, d[i], min_cov[i], dtype=np.int) for i in range(3)]

        # Get placement coordinate points
        placements = mgrid_to_points(np.meshgrid(*tuple(ds)))

        for p in placements:
            yield image.scaler.transform(X[p[0]:p[0]+self.dim,
                                         p[1]:p[1]+self.dim,
                                         p[2]:p[2]+self.dim]), p

    def get_patches_from(self, image, n_extra=0):
        for num, (p, coords) in enumerate(self.get_base_patches(image)):
            yield p, coords, "   Predicting on base patches (%i)" % (num+1)
        print("")
        for num, (p, coords) in enumerate(self.get_N_random_patches_from(image, n_extra)):
            yield p, coords, "   Predicting on extra patches (%i)" % (num+1)

    def validate_lab(self, lab, has_fg, cur_batch_size):
        valid = np.any(np.isin(self.fg_classes, lab))
        if valid:
            return valid, has_fg+1
        elif (self.n_fg_slices - has_fg) < (self.batch_size - cur_batch_size):
            # No FG, but there are still enough random slices left to fill the
            # minimum requirement
            return True, has_fg
        else:
            # No FG, but there is not enough random slices left to fill the
            # minimum requirement. Discard the slice and sample again.
            return False, has_fg

    def get_random_box_coords(self, im):
        dim = [max(0, s-self.dim) for s in im.shape[:3]]
        cords = np.round((dim * np.random.rand(3)).astype(np.uint16))
        return cords

    def get_box_coords(self, im):
        """
        Overwritten in SlidingPatchSequence3D to provide deterministic sampling
        """
        return self.get_random_box_coords(im)

    def __getitem__(self, idx, image_id=None):
        """
        Used by keras.fit_generator to fetch mini-batches during training
        """
        # If multiprocessing, set unique seed for this particular process
        self.seed()

        # Store how many slices has fg so far
        has_fg = 0

        # Interpolate on a random index for each sample image to generate batch
        batch_x, batch_y, batch_w = [], [], []

        with self.image_pair_loader.get_random() as image:
            while len(batch_x) < self.batch_size:

                # Fetch image, labels and weights
                X, y, w = image.image, image.labels, image.sample_weight

                # Sample a random box in the volume
                xc, yc, zc = self.get_box_coords(X)

                # Slice volume
                im = X[xc:xc+self.dim, yc:yc+self.dim, zc:zc+self.dim]
                lab = y[xc:xc+self.dim, yc:yc+self.dim, zc:zc+self.dim]

                # Make sure the box is of sufficient size
                im = center_expand(im, self.dim, self.bg_value, random=True)
                lab = center_expand(lab, self.dim, self.bg_value, random=True)

                # Validate label volume
                valid, has_fg = self.validate_lab(lab, has_fg, len(batch_y))
                if valid:
                    # Normalize image
                    im = image.scaler.transform(im)

                    batch_x.append(im)
                    batch_y.append(lab)
                    batch_w.append(w)

        # Reshape X (and one-hot encode y)
        batch_x = np.asarray(batch_x)
        batch_y = np.asarray(batch_y)
        batch_y = batch_y.reshape(batch_y.shape + (1,))

        return batch_x, batch_y, np.asarray(batch_w)

    def log(self):
        self.logger("Sequence Generator: %s" % self.__class__.__name__)
        self.logger("Box dimensions:     %s" % self.dim)
        self.logger("Batch size:         %s" % self.batch_size)
        self.logger("N fg slices/batch:  %s" % self.n_fg_slices)



import numpy as np
from mpunet.interpolation.regular_grid_interpolator import RegularGridInterpolator
from mpunet.logging import ScreenLogger
from mpunet.interpolation.linalg import points_to_mgrid, mgrid_to_points
from mpunet.interpolation.sample_grid import get_voxel_axes_real_space


def is_rot_mat(mat):
    """
    Validate that a square matrix is a rotation matrix
    """
    is_ortho = np.all(np.isclose(mat.dot(mat.T), mat.T.dot(mat)))
    is_unimodular = np.isclose(np.linalg.det(mat), 1)
    return is_ortho and is_unimodular


class ViewInterpolator(object):
    def __init__(self, image, labels, affine,
                 bg_value=0.0, bg_class=0, logger=None):

        # Ensure 4D
        if not image.ndim == 4:
            raise ValueError("Input img of dim %i must be dim 4."
                             "If image has only 1 channel, use "
                             "np.expand_dims(img, -1)." % image.ndim)

        # Set logger
        self.logger = logger if logger is not None else ScreenLogger()

        # Number of channels in the input image
        self.im_shape = image.shape
        self.n_channels = self.im_shape[-1]
        self.im_dtype = image.dtype

        # Cast bg-value to list of length n_channels
        if not isinstance(bg_value, (list, tuple, np.ndarray)):
            bg_value = [bg_value] * self.n_channels
        if not len(bg_value) == self.n_channels:
            raise ValueError("'bg_value' should be a list of length "
                             "'n_channels'. Got {} for n_channels={}".format(
                bg_value, self.n_channels))
        self.bg_value = bg_value

        # Store potential transformation to regular grid
        self.rot_mat = None

        # Define interpolators
        self.im_intrps, self.lab_intrp = self._init_interpolators(image,
                                                                  labels,
                                                                  bg_value,
                                                                  bg_class,
                                                                  affine)

    def apply_rotation(self, mgrid):
        if self.rot_mat is not None:
            shape = mgrid[0].shape
            rotated = self.rot_mat.dot(mgrid_to_points(mgrid).T).T
            return points_to_mgrid(rotated, shape)
        else:
            return mgrid

    def __call__(self, rgrid_mgrid):
        # Align grid if necessary
        rgrid_mgrid = self.apply_rotation(rgrid_mgrid)

        # Interpolate image and labels
        image = self.intrp_image(rgrid_mgrid, apply_rot=False)
        labels = self.intrp_labels(rgrid_mgrid, apply_rot=False)

        return image, labels

    def intrp_image(self, mgrid, apply_rot=True):
        if apply_rot:
            mgrid = self.apply_rotation(mgrid)

        if not isinstance(mgrid, tuple):
            # RegularGridInterpolator expects this tuple(xx, yy, zz) format
            mgrid = tuple(mgrid)

        # Interpolate image along all channels
        image = np.zeros(shape=mgrid[0].squeeze().shape + (self.n_channels,),
                         dtype=self.im_dtype)

        for i in range(self.n_channels):
            image[..., i] = self.im_intrps[i](mgrid).squeeze()

        return image

    def intrp_labels(self, mgrid, apply_rot=True):
        if apply_rot:
            mgrid = self.apply_rotation(mgrid)

        # Interpolate labels
        if self.lab_intrp:
            if not isinstance(mgrid, tuple):
                # RegularGridInterpolator expects tuple(xx, yy, zz) format
                mgrid = tuple(mgrid)

            return self._cast_labels(self.lab_intrp(mgrid).squeeze())
        else:
            return None

    def _init_interpolators(self, image, labels, bg_value, bg_class, affine):

        # Get voxel regular grid centered in real space
        g_all, basis, rot_mat = get_voxel_axes_real_space(image, affine,
                                                          return_basis=True)

        # Set rotation matrix
        self.rot_mat = rot_mat

        # Flip axes? Must be strictly increasing
        flip = np.sign(np.diagonal(basis)) == -1
        assert not np.any(flip)
        g_xx, g_yy, g_zz = g_all

        # Set interpolator for image, one for each channel
        im_intrps = []
        for i in range(self.n_channels):
            im_intrps.append(RegularGridInterpolator((g_xx, g_yy, g_zz),
                                                     image[..., i].squeeze(),
                                                     bounds_error=False,
                                                     fill_value=bg_value[i],
                                                     method="linear",
                                                     dtype=np.float32))

        try:
            # Set interpolator for labels
            lab_intrp = RegularGridInterpolator((g_xx, g_yy, g_zz), labels,
                                                bounds_error=False,
                                                fill_value=bg_class,
                                                method="nearest",
                                                dtype=np.uint8)
        except (AttributeError, TypeError, ValueError):
            lab_intrp = None

        return im_intrps, lab_intrp

    def _cast_labels(self, labels):
        # Cast labels float64 -> uint8/16
        type_info = np.iinfo(np.uint8)
        l = type_info.min
        u = type_info.max
        if np.all((labels >= l) & (labels <= u)):
            return labels.astype(np.uint8)
        else:
            return labels.astype(np.uint16)



import sqlite3
import numpy as np
from sqlite3 import Error
from pandas import DataFrame as df
import re

from mpunet.database import default_tables


class DBConnection(object):
    def __init__(self, db_file_path):
        self.db_file_path = db_file_path
        self._connection = None
        self._cursor = None
        self._connected = False

        # Boolean, return results from query directly
        self._echo = True
        self._auto_commit = True

        # Dict pointing to default table objects
        self.default_tables = {}
        self.create_default_table("DefaultTablesTable")

    def __repr__(self):
        """
        Developer string representation

        :return: str
        """
        return "DBConnection(%s)" % self.db_file_path

    def __str__(self):
        """
        Client string representation

        :return: str
        """
        s = "Database connection\n"
        s += "-"*len(s) + "\n"
        s += "DB:     %s\n" % self.db_file_path +\
             "Status: %s" % ("Connected" if self.connected else "Closed")
        return s

    def __call__(self, *args, **kwargs):
        """
        Delegates to self.query

        :param args: any arguments
        :param kwargs: any keyword arguments
        :return: return value of query method
        """
        if not self.connected:
            with self as conn:
                response = self.query(*args, **kwargs)
        else:
            response = self.query(*args, **kwargs)
        return response

    @property
    def echo(self):
        """
        Echo result set from query property getter

        :return: Boolean
        """
        return self._echo

    @echo.setter
    def echo(self, value):
        """
        Echo result set from query property setter

        :param value: boolean
        :return: None
        """
        if not isinstance(value, bool):
            raise ValueError("Echo must be True or False")
        self._echo = value

    @property
    def auto_commit(self):
        """
        Automatically commit all actions

        :return: Boolean
        """
        return self._auto_commit

    @auto_commit.setter
    def auto_commit(self, value):
        """
        Automatically commit all actions

        :param value: boolean
        :return: None
        """
        if not isinstance(value, bool):
            raise ValueError("Auto commit must be True or False")
        self._auto_commit = value

    def query(self, q_string, echo=None, to_numpy=True):
        """
        Executes a SQL query string against the database.
        Echoes the result back if echo=True or self.echo=True.

        :param q_string: SQL query string
        :param echo: echo back result, overwrites self.echo
        :return: SQL result set or Cursor object
        """
        words = q_string.upper().split(" ")
        if "DROP" in words:
            raise UserWarning("DROP queries should be performed manually.")

        # Define cursor object
        try:
            self.cursor = self.connection.cursor()
        except AttributeError:
            raise Error("Connection to DB has not been established.")

        # Execute query
        try:
            self.cursor.execute(q_string)
            if self.auto_commit:
                self.connection.commit()
        except Error as e:
            err_str = 'Error in query:\n"""\n%s\n"""' % q_string
            raise Error(err_str) from e

        # Fetch or return cursor
        echo = self.echo if echo is None else echo
        if echo and "SELECT" in words:
            if to_numpy:
                return np.array(self.cursor.fetchall())
            else:
                return self.cursor.fetchall()

    @property
    def connected(self):
        """
        Property getter for connection status (True/False)

        :return: boolean
        """
        return self._connected

    @property
    def connection(self):
        """
        Getter property for SQLite connection object

        :return: sqlite3.Connection or None
        """
        return self._connection

    @property
    def cursor(self):
        """
        Property getter for DB Cursor object

        :return: sqlite3.Cursor object
        """
        return self._cursor

    @cursor.setter
    def cursor(self, value):
        """
        Property setter for DB Cursor object

        :param value: sqlite.Cursor object
        :return: None
        """
        if not isinstance(value, sqlite3.Cursor):
            raise ValueError("cursor must be a sqlite3 Cursor object.")
        self._cursor = value

    @property
    def tables(self):
        """
        Return the names of all tables in the DB

        :return: list of strings of table names
        """
        with self as db:
            names = db.query("SELECT name FROM sqlite_master WHERE type='table';",
                             echo=True)
        return [n[0] for n in names]

    def print_table_info(self, table_name):
        """
        Return a list of column names for the table

        :param table_name: str, name of the table to fetch info on
        :return: list
        """
        if table_name not in self.tables:
            raise ValueError("No table '%s' in the DB." % table_name)

        tab_info = self("SELECT sql FROM sqlite_master "
                        "WHERE tbl_name = '%s' "
                        "AND type = 'table'" % table_name,
                        echo=True, to_numpy=False)[0][0].replace("\n", "")

        # Parse the info slightly
        columns = tab_info.split(f"{table_name}")[-1].strip(" (").split(",")
        print(f"Table name: {table_name}")
        for c in columns:
            name, type_ = c.split(" ", 1)
            print(name.ljust(25), type_)

    def create_default_table(self, default_table_name, **kwargs):
        # Create table object
        table = default_tables.__dict__[default_table_name](**kwargs)

        if default_table_name in self.tables:
            if default_table_name in self.default_tables:
                print("[OBS] Default table '%s' already exists" % default_table_name)
                return
        else:
            self(table.get_create_query(**kwargs))

        # Insert into dictionary
        self.default_tables[default_table_name] = table

        # For persistence, insert the table information into the
        # DefaultTablesTable from which the table object can be recreated at
        # a later time
        self.insert_into_default("DefaultTablesTable",
                                 table_name=default_table_name,
                                 **kwargs)

    def insert_into_default(self, default_table_name, **kwargs):
        # Get table and insert
        try:
            table = self.default_tables[default_table_name]
        except KeyError:
            # Recreate the table object
            table = None
        self(table.get_insert_query(**kwargs))

    def connect(self):
        """
        Attempts to connect to the SQLite database

        :return: None
        """
        try:
            self._connection = sqlite3.connect(self.db_file_path)
            self._connected = True

            # Enable foreign key support
            self.query("PRAGMA foreign_keys = ON;")

        except Error as e:
            raise Error("Could not connect to: %s" % self.db_file_path) from e
        return self

    def disconnect(self):
        """
        Disconnects from the SQLite database

        :return: None
        """
        self.connection.close()
        self._connected = False

    def __enter__(self):
        """
        On enter method for context manager usage. Establishes connection.

        :return: Connected DBConnection object
        """
        self.connect()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """
        On leave method for context manager usage. Disconnects the connection.
        Arguments are passed by Python automatically if an exception is raised
        within the context management block.

        :param exception_type: type of exception raised within context.
        :param exception_value: exception value from raised exception
        :param traceback: traceback of raised exception
        :return: None
        """
        self.disconnect()



"""
Prediction/evaluation script

Mathias Perslev
March 2018
"""

import os
import numpy as np
import nibabel as nib

from mpunet.utils.utils import (create_folders, get_best_model,
                                pred_to_class, await_PIDs)
from mpunet.logging.log_results import save_all
from mpunet.evaluate.metrics import dice_all
from argparse import ArgumentParser


def get_argparser():
    parser = ArgumentParser(description='Predict using a mpunet model.')
    parser.add_argument("--project_dir", type=str, default="./",
                        help='Path to mpunet project folder')
    parser.add_argument("-f", help="Predict on a single file")
    parser.add_argument("-l", help="Optional single label file to use with -f")
    parser.add_argument("--dataset", type=str, default="test",
                        help="Which dataset of those stored in the hparams "
                             "file the evaluation should be performed on. "
                             "Has no effect if a single file is specified "
                             "with -f.")
    parser.add_argument("--out_dir", type=str, default="predictions",
                        help="Output folder to store results")
    parser.add_argument("--num_GPUs", type=int, default=1,
                        help="Number of GPUs to use for this job")
    parser.add_argument("--sum_fusion", action="store_true",
                        help="Fuse the mutliple segmentation volumes into one"
                             " by summing over the probability axis instead "
                             "of applying a learned fusion model.")
    parser.add_argument("--overwrite", action='store_true',
                        help='Overwrite previous results at the output folder')
    parser.add_argument("--no_eval", action="store_true",
                        help="Perform no evaluation of the prediction performance. "
                             "No label files loaded when this flag applies.")
    parser.add_argument("--eval_prob", type=float, default=1.0,
                        help="Perform evaluation on only a fraction of the"
                             " computed views (to speed up run-time). OBS: "
                             "always performs evaluation on the combined "
                             "predictions.")
    parser.add_argument("--force_GPU", type=str, default="")
    parser.add_argument("--save_input_files", action="store_true",
                        help="Save in addition to the predicted volume the "
                             "input image and label files to the output dir)")
    parser.add_argument("--no_argmax", action="store_true",
                        help="Do not argmax prediction volume prior to save.")
    parser.add_argument("--on_val", action="store_true",
                        help="Evaluate on the validation set instead of test")
    parser.add_argument("--wait_for", type=str, default="",
                        help="Waiting for PID to terminate before starting "
                             "training process.")
    parser.add_argument("--continue", action="store_true",
                        help="Continue from a previsous, non-finished "
                             "prediction session at 'out_dir'.")
    return parser


def validate_folders(base_dir, out_dir, overwrite, _continue):
    """
    TODO
    """
    # Check base (model) dir contains required files
    must_exist = ("train_hparams.yaml", "views.npz",
                  "model")
    for p in must_exist:
        p = os.path.join(base_dir, p)
        if not os.path.exists(p):
            from sys import exit
            print("[*] Invalid mpunet project folder: '%s'"
                  "\n    Needed file/folder '%s' not found." % (base_dir, p))
            exit(0)

    # Check if output folder already exists
    if not (overwrite or _continue) and os.path.exists(out_dir):
        from sys import exit
        print("[*] Output directory already exists at: '%s'"
              "\n  Use --overwrite to overwrite or --continue to continue" % out_dir)
        exit(0)
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)


def save_nii_files(merged, image_pair, nii_res_dir, save_input_files):
    """
    TODO
    """
    # Extract data if nii files
    try:
        merged = merged.get_data()
    except AttributeError:
        merged = nib.Nifti1Image(merged, affine=image_pair.affine)
    volumes = [merged, image_pair.image_obj, image_pair.labels_obj]
    labels = ["%s_PRED.nii.gz" % image_pair.identifier,
              "%s_IMAGE.nii.gz" % image_pair.identifier,
              "%s_LABELS.nii.gz" % image_pair.identifier]
    if not save_input_files:
        volumes = volumes[:1]
        labels = labels[:1]
        p = os.path.abspath(nii_res_dir)  # Save file directly in nii_res_dir
    else:
        # Create sub-folder under nii_res_dir
        p = os.path.join(nii_res_dir, image_pair.identifier)
    create_folders(p)

    for nii, fname in zip(volumes, labels):
        try:
            nib.save(nii, "%s/%s" % (p, fname))
        except AttributeError:
            # No labels file?
            pass


def remove_already_predicted(all_images, out_dir):
    """
    TODO
    """
    nii_dir = os.path.join(out_dir, "nii_files")
    already_pred = [i.replace("_PRED", "").split(".")[0]
                    for i in filter(None, os.listdir(nii_dir))]
    print("[OBS] Not predicting on images: {} "
          "(--continue mode)".format(already_pred))
    return {k: v for k, v in all_images.items() if k not in already_pred}


def load_hparams(base_dir):
    """
    TODO
    """
    from mpunet.hyperparameters import YAMLHParams
    return YAMLHParams(os.path.join(base_dir, "train_hparams.yaml"))


def set_test_set(hparams, dataset):
    """
    TODO
    """
    hparams['test_dataset'] = hparams[dataset.strip("_dataset") + "_dataset"]


def set_gpu_vis(args):
    """
    TODO
    """
    force_gpu = args.force_GPU
    if not force_gpu:
        # Wait for free GPU
        from mpunet.utils import await_and_set_free_gpu
        await_and_set_free_gpu(N=args.num_GPUs, sleep_seconds=120)
        num_GPUs = args.num_GPUs
    else:
        from mpunet.utils import set_gpu
        set_gpu(force_gpu)
        num_GPUs = len(force_gpu.split(","))
    return num_GPUs


def get_image_pair_loader(args, hparams, out_dir):
    """
    TODO
    """
    from mpunet.image import ImagePairLoader, ImagePair
    if not args.f:
        # No single file was specified with -f flag, load the desired dataset
        dataset = args.dataset.replace("_data", "") + "_data"
        image_pair_loader = ImagePairLoader(predict_mode=args.no_eval,
                                            **hparams[dataset])
    else:
        predict_mode = not bool(args.l)
        image_pair_loader = ImagePairLoader(predict_mode=predict_mode,
                                            initialize_empty=True)
        image_pair_loader.add_image(ImagePair(args.f, args.l))

    # Put image pairs into a dict and remove from image_pair_loader to gain
    # more control with garbage collection
    image_pair_dict = {image.identifier: image for image in image_pair_loader.images}
    if vars(args)["continue"]:
        # Remove images that were already predicted
        image_pair_dict = remove_already_predicted(image_pair_dict, out_dir)
    return image_pair_loader, image_pair_dict


def get_results_dicts(out_dir, views, image_pairs_dict, n_classes, _continue):
    """
    TODO
    """
    from mpunet.logging import init_result_dicts, save_all, load_result_dicts
    if _continue:
        csv_dir = os.path.join(out_dir, "csv")
        results, detailed_res = load_result_dicts(csv_dir=csv_dir, views=views)
    else:
        # Prepare dictionary to store results in pd df
        results, detailed_res = init_result_dicts(views, image_pairs_dict, n_classes)
    # Save to check correct format
    save_all(results, detailed_res, out_dir)
    return results, detailed_res


def get_model(project_dir, build_hparams):
    """
    TODO
    """
    from mpunet.models.model_init import init_model
    model_path = get_best_model(project_dir + "/model")
    weights_name = os.path.splitext(os.path.split(model_path)[1])[0]
    print("\n[*] Loading model weights:\n", model_path)
    import tensorflow as tf
    with tf.distribute.MirroredStrategy().scope():
        model = init_model(build_hparams)
        model.load_weights(model_path, by_name=True)
    return model, weights_name


def get_fusion_model(n_views, n_classes, project_dir, weights_name):
    """
    TODO
    """
    from mpunet.models import FusionModel
    fm = FusionModel(n_inputs=n_views, n_classes=n_classes)
    # Load fusion weights
    weights = project_dir + "/model/fusion_weights/%s_fusion_" \
                            "weights.h5" % weights_name
    print("\n[*] Loading fusion model weights:\n", weights)
    fm.load_weights(weights)
    print("\nLoaded weights:\n\n%s\n%s\n---" % tuple(
        fm.layers[-1].get_weights()))
    return fm


def evaluate(pred, true, n_classes, ignore_zero=False):
    """
    TODO
    """
    pred = pred_to_class(pred, img_dims=3, has_batch_dim=False)
    return dice_all(y_true=true,
                    y_pred=pred,
                    ignore_zero=ignore_zero,
                    n_classes=n_classes,
                    skip_if_no_y=False)


def _per_view_evaluation(image_id, pred, true, mapped_pred, mapped_true, view,
                         n_classes, results, per_view_results, out_dir, args):
    """
    TODO
    """
    if np.random.rand() > args.eval_prob:
        print("Skipping evaluation for view %s... "
              "(eval_prob=%.3f)" % (view, args.eval_prob))
        return

    # Evaluate the raw view performance
    view_dices = evaluate(pred, true, n_classes)
    mapped_dices = evaluate(mapped_pred, mapped_true, n_classes)
    mean_dice = mapped_dices[~np.isnan(mapped_dices)][1:].mean()

    # Print dice scores
    print("View dice scores:   ", view_dices)
    print("Mapped dice scores: ", mapped_dices)
    print("Mean dice (n=%i): " % (len(mapped_dices) - 1), mean_dice)

    # Add to results
    results.loc[image_id, str(view)] = mean_dice
    per_view_results[str(view)][image_id] = mapped_dices[1:]

    # Overwrite with so-far results
    save_all(results, per_view_results, out_dir)


def _merged_eval(image_id, pred, true, n_classes, results,
                 per_view_results, out_dir):
    """
    TODO
    """
    # Calculate combined prediction dice
    dices = evaluate(pred, true, n_classes, ignore_zero=True)
    mean_dice = dices[~np.isnan(dices)].mean()
    per_view_results["MJ"][image_id] = dices

    print("Combined dices: ", dices)
    print("Combined mean dice: ", mean_dice)
    results.loc[image_id, "MJ"] = mean_dice

    # Overwrite with so-far results
    save_all(results, per_view_results, out_dir)


def _multi_view_predict_on(image_pair, seq, model, views, results,
                           per_view_results, out_dir, args):
    """
    TODO
    """
    from mpunet.utils.fusion import predict_volume, map_real_space_pred
    from mpunet.interpolation.sample_grid import get_voxel_grid_real_space

    # Get voxel grid in real space
    voxel_grid_real_space = get_voxel_grid_real_space(image_pair)

    # Prepare tensor to store combined prediction
    d = image_pair.image.shape[:-1]
    combined = np.empty(
        shape=(len(views), d[0], d[1], d[2], seq.n_classes),
        dtype=np.float32
    )
    print("Predicting on brain hyper-volume of shape:", combined.shape)

    # Predict for each view
    for n_view, view in enumerate(views):
        print("\n[*] (%i/%i) View: %s" % (n_view + 1, len(views), view))
        # for each view, predict on all voxels and map the predictions
        # back into the original coordinate system

        # Sample planes from the image at grid_real_space grid
        # in real space (scanner RAS) coordinates.
        X, y, grid, inv_basis = seq.get_view_from(image_pair, view,
                                                  n_planes="same+20")

        # Predict on volume using model
        pred = predict_volume(model, X, axis=2, batch_size=seq.batch_size)

        # Map the real space coordiante predictions to nearest
        # real space coordinates defined on voxel grid
        mapped_pred = map_real_space_pred(pred, grid, inv_basis,
                                          voxel_grid_real_space,
                                          method="nearest")
        combined[n_view] = mapped_pred

        if not args.no_eval:
            _per_view_evaluation(image_id=image_pair.identifier,
                                 pred=pred,
                                 true=y,
                                 mapped_pred=mapped_pred,
                                 mapped_true=image_pair.labels,
                                 view=view,
                                 n_classes=seq.n_classes,
                                 results=results,
                                 per_view_results=per_view_results,
                                 out_dir=out_dir,
                                 args=args)
    return combined


def merge_multi_view_preds(multi_view_preds, fusion_model, args):
    """
    TODO
    """
    fm = fusion_model
    if not args.sum_fusion:
        # Combine predictions across views using Fusion model
        print("\nFusing views (fusion model)...")
        d = multi_view_preds.shape
        multi_view_preds = np.moveaxis(multi_view_preds, 0, -2)
        multi_view_preds = multi_view_preds.reshape((-1, fm.n_inputs, fm.n_classes))
        merged = fm.predict(multi_view_preds, batch_size=10**4, verbose=1)
        merged = merged.reshape((d[1], d[2], d[3], fm.n_classes))
    else:
        print("\nFusion views (sum)...")
        merged = np.sum(multi_view_preds, axis=0)
    merged_map = pred_to_class(merged.squeeze(), img_dims=3).astype(np.uint8)
    return merged, merged_map


def run_predictions_and_eval(image_pair_loader, image_pair_dict, model,
                             fusion_model, views, hparams, args, results,
                             per_view_results, out_dir, nii_res_dir):
    """
    TODO
    """
    # Set scaler and bg values
    image_pair_loader.set_scaler_and_bg_values(
        bg_value=hparams.get_from_anywhere('bg_value'),
        scaler=hparams.get_from_anywhere('scaler'),
        compute_now=False
    )

    # Init LazyQueue and get its sequencer
    from mpunet.sequences.utils import get_sequence
    seq = get_sequence(data_queue=image_pair_loader,
                       is_validation=True,
                       views=views,
                       **hparams["fit"], **hparams["build"])

    from mpunet.utils.fusion.fuse_and_predict import predict_single
    o = predict_single(image_pair_loader.images[0], model, hparams)
    print(o.shape)

    image_ids = sorted(image_pair_dict)
    n_images = len(image_ids)
    for n_image, image_id in enumerate(image_ids):
        print("\n[*] (%i/%s) Running on: %s" % (n_image + 1, n_images, image_id))

        with seq.image_pair_queue.get_image_by_id(image_id) as image_pair:
            # Get prediction through all views
            multi_view_preds = _multi_view_predict_on(
                image_pair=image_pair,
                seq=seq,
                model=model,
                views=views,
                results=results,
                per_view_results=per_view_results,
                out_dir=out_dir,
                args=args
            )

            # Merge the multi view predictions into a final segmentation
            merged, merged_map = merge_multi_view_preds(multi_view_preds,
                                                        fusion_model, args)
            if not args.no_eval:
                _merged_eval(
                    image_id=image_id,
                    pred=merged_map,
                    true=image_pair.labels,
                    n_classes=hparams["build"]["n_classes"],
                    results=results,
                    per_view_results=per_view_results,
                    out_dir=out_dir
                )

            # Save combined prediction volume as .nii file
            print("Saving .nii files...")
            save_nii_files(merged=merged_map if not args.no_argmax else merged,
                           image_pair=image_pair,
                           nii_res_dir=nii_res_dir,
                           save_input_files=args.save_input_files)


def assert_args(args):
    pass


def entry_func(args=None):
    # Get command line arguments
    args = get_argparser().parse_args(args)
    assert_args(args)

    # Get most important paths
    project_dir = os.path.abspath(args.project_dir)
    out_dir = os.path.abspath(args.out_dir)

    # Check if valid dir structures
    validate_folders(project_dir, out_dir,
                     overwrite=args.overwrite,
                     _continue=vars(args)["continue"])
    nii_res_dir = os.path.join(out_dir, "nii_files")
    create_folders(nii_res_dir, create_deep=True)

    # Get settings from YAML file
    hparams = load_hparams(project_dir)

    # Get dataset
    image_pair_loader, image_pair_dict = get_image_pair_loader(args, hparams,
                                                               out_dir)

    # Wait for PID to terminate before continuing, if specified
    if args.wait_for:
        await_PIDs(args.wait_for, check_every=120)

    # Set GPU device
    set_gpu_vis(args)

    # Get views
    views = np.load("%s/views.npz" % project_dir)["arr_0"]
    del hparams['fit']['views']

    # Prepare result dicts
    results, per_view_results = None, None
    if not args.no_eval:
        results, per_view_results = get_results_dicts(out_dir, views,
                                                      image_pair_dict,
                                                      hparams["build"]["n_classes"],
                                                      vars(args)["continue"])

    # Get model and load weights, assign to one or more GPUs
    model, weights_name = get_model(project_dir, hparams['build'])
    fusion_model = None
    if not args.sum_fusion:
        fusion_model = get_fusion_model(n_views=len(views),
                                        n_classes=hparams["build"]["n_classes"],
                                        project_dir=project_dir,
                                        weights_name=weights_name)

    run_predictions_and_eval(
        image_pair_loader=image_pair_loader,
        image_pair_dict=image_pair_dict,
        model=model,
        fusion_model=fusion_model,
        views=views,
        hparams=hparams,
        args=args,
        results=results,
        per_view_results=per_view_results,
        out_dir=out_dir,
        nii_res_dir=nii_res_dir
    )
    if not args.no_eval:
        # Write final results
        save_all(results, per_view_results, out_dir)


if __name__ == "__main__":
    entry_func()



import os
from argparse import ArgumentParser


def copy_yaml_and_set_data_dirs(in_path, out_path, data_dir):
    from mpunet.hyperparameters import YAMLHParams
    hparams = YAMLHParams(in_path, no_log=True, no_version_control=True)
    dir_name = "base_dir" if "base_dir" in hparams["train_data"] else "data_dir"

    # Set values in parameter file and save to new location
    data_ids = ("train", "val", "test") + (("aug",)
                                           if hparams.get("aug_data")
                                           else ())
    for dataset in data_ids:
        path = (data_dir + "/{}".format(dataset)) if data_dir else "Null"
        dataset = dataset + "_data"
        if not hparams.get(dataset) or not hparams[dataset].get("base_dir"):
            try:
                hparams.set_value(dataset, dir_name, path, True, True)
            except AttributeError:
                print("[!] Subdir {} does not exist.".format(dataset))
    hparams.save_current(out_path)


def get_parser():
    parser = ArgumentParser(description='Create a new project folder')

    # Define groups
    parser._action_groups.pop()
    required = parser.add_argument_group('required named arguments')
    optional = parser.add_argument_group('optional named arguments')

    models = [m for m in os.listdir(os.path.split(__file__)[0] + "/defaults")
              if m != "__init__.py"]
    required.add_argument('--name', type=str, required=True,
                        help='the name of the project folder')
    optional.add_argument('--root', type=str, default=os.path.abspath("./"),
                          help='a path to the root folder in '
                               'which the project will be initialized')
    optional.add_argument("--model", type=str, default="MultiPlanar",
                          help="Specify a model type parameter file "
                               "({}), default="
                               "'MultiPlanar'".format(", ".join(models)))
    optional.add_argument("--data_dir", type=str, default=None,
                          help="Root data folder for the project")
    return parser


def entry_func(args=None):

    default_folder = os.path.split(os.path.abspath(__file__))[0] + "/defaults"
    if not os.path.exists(default_folder):
        raise OSError("Default path not found at %s" % default_folder)

    # Parse arguments
    parser = get_parser()
    args = vars(parser.parse_args(args))
    path = os.path.abspath(args["root"])
    name = args["name"]
    preset = args["model"]
    data_dir = args["data_dir"]
    if data_dir:
        data_dir = os.path.abspath(data_dir)

    # Validate project path and create folder
    if not os.path.exists(path):
        raise OSError("root path '%s' does not exist." % args["root"])
    else:
        folder_path = "%s/%s" % (path, name)
        if os.path.exists(folder_path):
            response = input("Folder at '%s' already exists. Overwrite? "
                             "Only parameter files and code will be replaced. (y/n) " % folder_path)
            if response.lower() == "n":
                raise OSError("Folder at '%s' already exists" % folder_path)
        else:
            os.makedirs("%s/%s" % (path, name))

    # Get yaml path
    from glob import glob
    yaml_paths = glob(os.path.join(default_folder, preset, "*.yaml"))

    # Write file
    for p in yaml_paths:
        copy_yaml_and_set_data_dirs(in_path=p,
                                    out_path=os.path.join(folder_path, os.path.split(p)[-1]),
                                    data_dir=data_dir)


if __name__ == "__main__":
    entry_func()
